/*
 * $Id: cia-tmpl.c,v 1.1 1997/05/22 19:40:10 fachat Exp $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * Written by
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Fixes by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 * $Log: cia-tmpl.c,v $
 * Revision 1.1  1997/05/22 19:40:10  fachat
 * Initial revision
 *
 */

/* 
 * 15mar97 a.fachat
 * new interrupt handling, hopefully according to the specs now.
 * All interrupts (note: not timer events (i.e. alarms)) are put
 * into one interrupt flag, I_MYCIAFL.
 * if an interrupt condition changes, the function (i.e. cpp macro)
 * update_myciairq() id called, that checks the IRQ line state.
 * This is now possible, as ettore has decoupled A_* alarm events
 * from I_* interrupts for performance reasons.
 */


#define	OLDIRQ

#define	myciaicr	mycia[CIA_ICR]

#define update_myciairq() 						\
	do { 								\
	  if( myciaint & myciaicr & 0x7f ) myciaint |= 0x80; 		\
    	  maincpu_set_int(I_MYCIAFL, myciaint & 0x80 ? MYCIA_INT : 0); 	\
	} while(0)

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "extern.h"
#include "vmachine.h"
#include "extern.h"
#include "cpu.h"
#include "cia.h"

INCLUDES
	
#include "interrupt.h"

/* #define MYCIA_TIMER_DEBUG */

#define	STORE_OFFSET 1 /* store_cia called one cycle after actual store */
#define	INT_OFFSET   1 /* offset given to int_cia is this amount to large */
#define	IRQ_OFFSET   4 /* IRQ line goes this # cycles low after ICR flag set */

/* global */

BYTE mycia[16];
extern BYTE keyarr[KBD_ROWS];

/* local functions */

static void update_myciata ( int );
static void update_myciatb ( int );
static void check_myciatodalarm( void );
int myciatbx ( int );

/*
 * Local variables
 */


static int myciaint;	/* Interrupt Flag register for cia 1 */

static int myciata;	/* value of mycia timer A at last update */
static int myciatb;	/* value of mycia timer B at last update */

static CLOCK rclk;	/* real clock = clk-offset */
static CLOCK myciatau;	/* time when mycia timer A is updated */
static CLOCK myciatai;	/* clock when next IRQ is expected */
static CLOCK myciatairq;/* clk when to set IRQ line low */
static CLOCK lastmyciatai;
static CLOCK myciatbu;	/* time when mycia timer B is updated */
static CLOCK myciatbi;	/* clock when next IRQ is expected */
static CLOCK myciatbirq;/* clk when to set IRQ line low */
static CLOCK lastmyciatbi;

static CLOCK myciardi;	/* time of last ICR read */

static int myciata_stopped = 1;	/* timer stopped */
static int myciatb_stopped = 1;	/* timer stopped */

static int myciasr_bits;	/* number of bits still to send * 2 */

static BYTE myciatodalarm[4];
static BYTE myciatodlatch[4];
static char myciatodstopped;
static char myciatodlatched;
static int myciatodticks = 100000;          /* approx. a 1/10 sec. */

/* ------------------------------------------------------------------------- */
/* MYCIA */

PRE_CIA_FUNCS

static int readta(void) {
	register int tmp;

	if(myciata_stopped) 
	  return (myciata==0) ? mycia[CIA_TAL]+256*mycia[CIA_TAH] : myciata; 

	tmp= (clk-myciatau + (myciata+1)) % (myciata+1);

	if(tmp==myciata) tmp = myciata;
	else tmp = myciata - tmp;

	return tmp;
}

static int readtb(void) {
	register int tmp;

	if(myciatb_stopped) 
	  return (myciatb==0) ? mycia[CIA_TBL]+256*mycia[CIA_TBH] : myciatb; 

	tmp= (clk-myciatbu + (myciatb+1)) % (myciatb+1);

	if(tmp==myciatb) tmp = myciatb;
	else tmp = myciatb - tmp;

	return tmp;
}


void    reset_mycia(void)
{
	int i;
	for(i=0;i<16;i++) mycia[i] = 0;

	/* FIXME: Is this actually correct? */
	mycia[CIA_TAL] = mycia[CIA_TBL] = 1;
	
	myciasr_bits = 0;
	myciata_stopped = 0;
	myciatb_stopped = 0;
	myciata = 0xffff;
	myciatb = 0xffff;
	myciatau = clk;
	myciatbu = clk;

	myciatbi = 0;
	myciatbirq = 0;
	lastmyciatbi = 0;
	maincpu_unset_alarm(A_MYCIATB);
	myciatai = 0;
	myciatairq = 0;
	lastmyciatai = 0;
	maincpu_unset_alarm(A_MYCIATA);

	myciardi = 0;

        memset(myciatodalarm, 0, sizeof(myciatodalarm));
        myciatodlatched = 0;
        myciatodstopped = 0;
        maincpu_set_alarm(A_MYCIATOD, myciatodticks);

	myciaint = 0;
	update_myciairq();
}


void    store_mycia(ADDRESS addr, BYTE byte)
{
    addr &= 0xf;
       
 
#ifdef DEBUG
    if (addr >=4 /*debugflg*/)
	printf("store mycia[%x] %x\n", (int) addr, (int) byte);
#endif

    if(myciatbi && myciatbi<clk-STORE_OFFSET) int_myciatb(clk-myciatbi);
    if(myciatai && myciatai<clk-STORE_OFFSET) int_myciata(clk-myciatai);

    rclk = clk - STORE_OFFSET;

    switch (addr) {

      case CIA_PRA: /* port A */
	STORE_SERIAL_CIAPA
      case CIA_DDRA: 
	STORE_CIAPA
	break;
      case CIA_PRB: /* port B */
      case CIA_DDRB: 
	STORE_CIAPB
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 *
	 * REAL:  TOD register + (wallclock - myciatodrel)
	 * VIRT:  TOD register + (cycles - begin)/cycles_per_sec
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
        if(mycia[CIA_CRB] & 0x80) myciatodalarm[addr-CIA_TOD_TEN]=byte;
        else {
          if(addr == CIA_TOD_TEN) myciatodstopped = 0;
          if(addr == CIA_TOD_HR) myciatodstopped = 1;
          mycia[addr] = byte;
        }
	check_myciatodalarm();
        break;

      case CIA_SDR: /* Serial Port output buffer */
	mycia[addr] = byte;
	if((mycia[CIA_CRA] & 0x40) && (myciasr_bits < 16)) myciasr_bits += 16;
	break;

	/* Interrupts */

      case CIA_ICR: /* Interrupt Control Register */
#if defined (MYCIA_TIMER_DEBUG)
        if(debugflg) printf ("MYCIA set CIA_ICR: 0x%x\n", byte);
#endif
	if(byte & CIA_IM_SET) {
	  mycia[addr] |= byte;
	} else {
	  mycia[addr] &= ~byte;
	}
#ifdef OLDIRQ
	maincpu_set_int(I_MYCIATA, (myciaint & mycia[CIA_ICR] & CIA_IM_TA) ? MYCIA_INT : 0 );
	if(!(myciaint & mycia[CIA_ICR] & CIA_IM_TA)) myciatairq = 0;
	if((!myciatairq) || (myciatai < myciatairq)) maincpu_set_alarm_clk(A_MYCIATA, myciatai);
	maincpu_set_int(I_MYCIATB, (myciaint & mycia[CIA_ICR] & CIA_IM_TB) ? MYCIA_INT : 0 );
	if(!(myciaint & mycia[CIA_ICR] & CIA_IM_TB)) myciatbirq = 0;
	/* TODO: this is a bad hack (tm)! */
	if((!(mycia[CIA_CRB] & 0x60)) && ((!myciatbirq) || (myciatbi < myciatbirq))) maincpu_set_alarm_clk(A_MYCIATB, myciatbi);
	maincpu_set_int(I_MYCIASP, (myciaint & mycia[CIA_ICR] & CIA_IM_SP) ? MYCIA_INT : 0 );
	maincpu_set_int(I_MYCIATOD,(myciaint & mycia[CIA_ICR] & CIA_IM_TOD)? MYCIA_INT : 0 );
	/* EP */
	if (myciaint & mycia[CIA_ICR])
	    myciaint |= 0x80;
	break;
#else
	update_myciairq();
	if(!(myciaint & mycia[CIA_ICR] & CIA_IM_TA)) myciatairq = 0;
	if((!myciatairq) || (myciatai < myciatairq)) maincpu_set_alarm_clk(A_MYCIATA, myciatai);
	if(!(myciaint & mycia[CIA_ICR] & CIA_IM_TB)) myciatbirq = 0;
	/* TODO: this is a bad hack (tm)! */
	if((!(mycia[CIA_CRB] & 0x60)) && ((!myciatbirq) || (myciatbi < myciatbirq))) maincpu_set_alarm_clk(A_MYCIATB, myciatbi);
#endif

	/* Control */

      case CIA_CRA: /* control register A */
#if defined (MYCIA_TIMER_DEBUG)
        printf ("MYCIA set CIA_CRA: 0x%x\n", byte);
#endif

	/* bit 7 tod frequency */
	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_myciata(STORE_OFFSET + INT_OFFSET -1);
	    if(!myciata_stopped) {
	      	myciatau = clk;
	      	myciatai = clk + myciata;
		if((!myciatairq) || (myciatairq>myciatai)) {
		    maincpu_set_alarm(A_MYCIATA, myciata);
		}
	    }
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
            /* stop timer */
#if defined (MYCIA_TIMER_DEBUG)
	    if(debugflg) printf ("MYCIA Timer A stopped\n");
#endif
	    if(!myciata_stopped) {
	    	myciata = readta(); 	/* set myciata to actual value */
		myciata_stopped = 1;
		myciatai = 0;
		if(myciatairq) {
		    maincpu_set_alarm_clk(A_MYCIATA, myciatairq);
		} else {
		    maincpu_unset_alarm(A_MYCIATA);
		}
	    }
	} else {
#if defined (MYCIA_TIMER_DEBUG)
	    printf ("MYCIA Timer A restarted, myciata=%d\n", myciata);
#endif
	    if(myciata_stopped) {
		myciatau = clk;
		myciata_stopped = 0;
                update_myciata(STORE_OFFSET + INT_OFFSET -1); 
		myciatai = clk + myciata;
		if((!myciatairq) || (myciatairq>myciatai))
		    maincpu_set_alarm(A_MYCIATA, myciata);
	    }
	}

	mycia[addr] = byte & 0xef;	/* remove strobe */
	break;

      case CIA_CRB: /* control register B */
#if defined (MYCIA_TIMER_DEBUG)
        if(1 /*debugflg*/) printf ("MYCIA set CIA_CRB: 0x%x, stopped=%d\n", 
				byte, myciatb_stopped);
#endif
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_myciatb(STORE_OFFSET + INT_OFFSET -1);
	    if(!myciatb_stopped) {
	      	myciatbu = clk;
	      	myciatbi = clk + myciatb;
		if((!myciatbirq) || (myciatbirq>myciatbi))
		    maincpu_set_alarm(A_MYCIATB, myciatb);
	    }
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x60) || !(byte & 1)) {
	    if(!myciatb_stopped) {
#if defined (MYCIA_TIMER_DEBUG)
 	    	if(debugflg) printf ("MYCIA Timer B stopped\n");
#endif
		myciatb = readtb(); 	/* set myciatb to actual value */
		myciatb_stopped = 1;
		myciatbi = 0;
		if(myciatbirq) 
		   maincpu_set_alarm_clk(A_MYCIATB, myciatbirq);
		else
		   maincpu_unset_alarm(A_MYCIATB);
	    }
	} else {
	    if(myciatb_stopped) {
#if defined (MYCIA_TIMER_DEBUG)
	    	if(debugflg) printf ("MYCIA Timer B restarted\n");
#endif
		myciatbu = clk;
		myciatb_stopped = 0;
                update_myciatb(STORE_OFFSET + INT_OFFSET -1);
		myciatbi = clk + myciatb;
/*printf("start myciatb: clk = %ld, myciatb =%04x, myciatbirq =%ld, myciatbi=%ld\n",
			clk, myciatb, myciatbirq, myciatbi);*/
		if((!myciatbirq) || (myciatbirq>myciatbi))
		    maincpu_set_alarm(A_MYCIATB, myciatb);
	    }
	}

	mycia[addr] = byte & 0xef;	/* remove strobe */
	break;
      default:
	mycia[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_mycia(ADDRESS addr)
{
    addr &= 0xf;
    
#ifdef DEBUG
    if (addr < 4 /*debugflg*/) printf("read mycia[%x], clk=%ld\n", addr, clk);
#endif
   
/*    if(addr == CIA_ICR) myciardi = clk;*/
 
    if(myciatbi && myciatbi<clk) int_myciatb(clk-myciatbi);
    if(myciatai && myciatai<clk) int_myciata(clk-myciatai);

    switch (addr) {

      case CIA_PRA: /* port A */
	READ_CIAPA
	break;

      case CIA_PRB: /* port B */
	READ_CIAPB
	break;
	/* Timers */

      case CIA_TAL: /* timer A low */
	return (readta() & 0xff);

      case CIA_TAH: /* timer A high */
	return ((readta() >> 8) & 0xff);

      case CIA_TBL: /* timer B low */
	return readtb() & 0xff;

      case CIA_TBH: /* timer B high */
	return (readtb() >> 8) & 0xff;

	/*
	 * TOD clock is latched by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this latching is different from the input one.
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
        if(!myciatodlatched) memcpy(myciatodlatch, mycia+CIA_TOD_TEN, sizeof(myciatodlatch));
        if(addr == CIA_TOD_TEN) myciatodlatched = 0;
        if(addr == CIA_TOD_HR) myciatodlatched = 1;
        return mycia[addr];

      case CIA_SDR: /* Serial Port Shift Register */
#ifdef DEBUG
	/* little hack .... */
	{ int i; printf("\nmaincpu_ints:");
	for(i=0;i<NUMOFINT;i++) {
	  printf(" %d",maincpu_int_status.pending_int[i]);
	}
	printf("\n"); }
#endif
	return (mycia[addr]);

	/* Interrupts */

      case CIA_ICR: /* Interrupt Flag Register */
	{
	    BYTE    t = myciaint;

	    myciardi = clk;
	    /* if ( t & mycia[CIA_ICR] ) t |= 0x80; EP (removed) */

#ifdef DEBUG
	    if(debugflg) 
		printf("MYCIA read intfl gives myciaint=%02x -> %02x @"
		   	" PC = %04x, sr_bits=%d, clk=%ld, ta=%d, tb=%d\n", 
			myciaint, t, PC, myciasr_bits, clk, readta(), readtb());
#endif
	    myciaint = 0;
	    myciatairq = 0;
	    myciatbirq = 0;
	    if(myciatai) maincpu_set_alarm_clk(A_MYCIATA, myciatai);
	    if(myciatbi) maincpu_set_alarm_clk(A_MYCIATB, myciatbi);
#ifdef OLDIRQ	    
	    maincpu_set_int(I_MYCIATA, 0);
	    maincpu_set_int(I_MYCIATB, 0);
	    maincpu_set_int(I_MYCIASP, 0);
	    maincpu_set_int(I_MYCIATOD,0);
#else
	    update_myciairq();
#endif
	    return (t);
	}
    }  /* switch */

    return (mycia[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_myciata(long offset)
{

    if(offset<0) return 0; 	/* ???????? */

    rclk = clk - offset;

#ifdef DEBUG
    if (debugflg)
	printf("mycia timer A event\n");
#endif

#ifdef MYCIA_TIMER_DEBUG
printf("int_myciata: rclk=%ld-%ld=%ld: myciatai=%ld, lastatai=%ld, tau=%ld, tairq=%ld\n",
			clk, offset, clk-offset,myciatai,lastmyciatai, myciatau,
			myciatairq);
#endif

    if(rclk == myciatairq) { 
#ifdef OLDIRQ
      if(mycia[CIA_ICR] & CIA_IM_TA) {
	maincpu_set_int(I_MYCIATA, MYCIA_INT);
	myciaint |= 0x80;	/* EP */
      }
#else
      myciaint |= CIA_IM_TA;
      update_myciairq();
#endif
      myciatairq = 0;
      if(myciatai) {

#ifdef MYCIA_TIMER_DEBUG
printf("int_myciata: maincpu_set_alarm(%ld -> %ld)\n",myciatai - clk, myciatai);
#endif
	maincpu_set_alarm_clk(A_MYCIATA, myciatai);
      } else {
	maincpu_unset_alarm(A_MYCIATA);
      }
    } 

    if(rclk == myciatai) {

      lastmyciatai = myciatai;

      /* update timer */

      if (mycia[CIA_CRA] & 8) {	/* one-shot mode */
#if defined (MYCIA_TIMER_DEBUG)
        if(debugflg) printf ("MYCIA Timer A event -- one-shot mode: next int won't happen\n");
#endif
	myciata = mycia[CIA_TAL]+256*mycia[CIA_TAH];
	myciata_stopped = 1;
	myciatai = 0;
	if(myciatairq) 
	  maincpu_set_alarm_clk(A_MYCIATA, myciatairq);
	else
	  maincpu_unset_alarm(A_MYCIATA);
	mycia[CIA_CRA] &= ~1;
      } else {		/* continuous mode */
	update_myciata(offset);
      }

      /* check SR */

      if(mycia[CIA_CRA] & 0x40) {
        if(myciasr_bits) {
  	  if(!(--myciasr_bits)) {
	    myciaint |= CIA_IM_SP;
#ifdef OLDIRQ
	    /* TODO: this sure is another delayed IRQ... */
            if(mycia[CIA_ICR] & CIA_IM_SP) {
	      maincpu_set_int(I_MYCIASP, MYCIA_INT);
	      myciaint |= 0x80;	/* EP */
	    }
#else
	    update_myciairq();
#endif
	  }
        } 
      }

      /* set IRQ line (delayed, if necessary) */

      if(1 /*myciardi != rclk-STORE_OFFSET*/) {	/* not cleared by reading ICR */
#ifdef OLDIRQ
        myciaint |= CIA_IM_TA;
#endif
        if(mycia[CIA_ICR] & CIA_IM_TA) {  /* timer bitmask set? */
#ifdef MYCIA_TIMER_DEBUG
printf("clk=%ld: lastmyciatai=%ld\n",clk, lastmyciatai);
#endif
	  if((clk-lastmyciatai) > IRQ_OFFSET) {
#ifdef OLDIRQ
	    maincpu_set_int(I_MYCIATA, MYCIA_INT);
	    myciaint |= 0x80;	/* EP */
#else
	    myciaint |= CIA_IM_TA;
	    update_myciairq();
#endif
	    myciatairq = 0;
	  } else {
#ifdef MYCIA_TIMER_DEBUG
printf("int_myciata: main_cpu_set_alarm(%ld -> %ld)\n", lastmyciatai - clk + IRQ_OFFSET,
			lastmyciatai + IRQ_OFFSET);
#endif
#ifdef OLDIRQ
	    myciaint |= 0x80;
#endif
	    myciatairq = lastmyciatai + IRQ_OFFSET;
	    if((!myciatai) || myciatairq < myciatai) {
	      maincpu_set_alarm_clk(A_MYCIATA, myciatairq);
	    } else {
	      maincpu_set_alarm_clk(A_MYCIATA, myciatai);
	    }
	  }
	}
      }

    }
      /* count timer A underflow pulses */

      if ((mycia[CIA_CRB] & 0x61) == 0x41) {
	if (--myciatb < 0) {	/* is this "=" ok? (AF) */
	    myciatbx(offset); /*int_myciatb(0); */
	}
      }

    return 0; 
}


/*
 * Timer B can run in 2 (4) modes
 * mycia[f] & 0x60 == 0x00   count system 02 pulses
 * mycia[f] & 0x60 == 0x40   count timer A underflows
 * mycia[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


int    int_myciatb(long offset)
{

    if(offset<0) {
	return 0; 	/* ???????? */
    }

    rclk = clk - offset;

#ifdef DEBUG
    if (debugflg)
	printf("MYCIA timer B event, clk=%ld, offset=%ld, tbi=%ld, latch=%d,"
		"PC=%04x, rline=%d\n",
		 clk, offset, myciatbi, mycia[CIA_TBL]+256*mycia[CIA_TBH],
		PC, rasterline);
#endif

#ifdef MYCIA_TIMER_DEBUG
printf("int_myciatb: clk=%ld-%ld: myciatbi=%ld\n",clk, offset,myciatbi);
#endif

/*    lastmyciatbi = myciatbirq;*/

    if(rclk == myciatbirq) {
#ifdef OLDIRQ
        if(mycia[CIA_ICR] & CIA_IM_TB) {
	  maincpu_set_int(I_MYCIATB, MYCIA_INT);
	  myciaint |= 0x80;	/* EP */
	}
#else
	myciaint |= CIA_IM_TB;
	update_myciairq();
#endif
	myciatbirq = 0;

	if(myciatbi) {
#ifdef MYCIA_TIMER_DEBUG
printf("int-myciatb: maincpu_set_alarm(%ld)\n",myciatbi - clk);
#endif
	  maincpu_set_alarm_clk(A_MYCIATB, myciatbi);
	} else {
	  maincpu_unset_alarm(A_MYCIATB);
	}
    } 

    /* FIXME: hack to allow int_myciatb() to be called from int_myciata() */
    if(rclk == myciatbi) {
	return myciatbx(offset);
    }

    return 0;
}

int myciatbx(int offset) {

      lastmyciatbi = rclk; /*myciatbi;*/
/*printf("tbx: clk=%ld, lastmyciatbi=%ld, icr=%02x\n", clk, lastmyciatbi, mycia[CIA_ICR]);*/

      /* update timer */

      if (mycia[CIA_CRB] & 8) {	/* one-shot mode */
#if defined (MYCIA_TIMER_DEBUG)
        printf ("MYCIA Timer B event -- one-shot mode: next int won't happen\n");
#endif
	myciatb = mycia[CIA_TBL]+256*mycia[CIA_TBH];
	myciatb_stopped = 1;
	myciatbi = 0;
	if(myciatbirq) 
	  maincpu_set_alarm_clk(A_MYCIATB, myciatbirq);
	else
	  maincpu_unset_alarm(A_MYCIATB);
	mycia[CIA_CRB] &= ~1;	/* CRB run bit */
      } else {		/* continuous mode */
	update_myciatb(offset);
      }

      /* set IRQ line (delayed, if necessary) */

      if(1 /*myciardi != rclk-STORE_OFFSET*/) {
#ifdef OLDIRQ
        myciaint |= CIA_IM_TB;
#endif
        if(mycia[CIA_ICR] & CIA_IM_TB) {
#ifdef MYCIA_TIMER_DEBUG
printf("clk=%ld: lastmyciatbi=%ld\n",clk, lastmyciatbi);
#endif
	  if((clk-lastmyciatbi) > IRQ_OFFSET) {
#ifdef OLDIRQ
	    maincpu_set_int(I_MYCIATB, MYCIA_INT);
	    myciaint |= 0x80;	/* EP */
#else
	    myciaint |= CIA_IM_TB;
	    update_myciairq();
#endif
	    myciatbirq = 0;
	  } else {
#ifdef MYCIA_TIMER_DEBUG
printf("int_myciatb: maincpu_set_alarm(%ld -> %ld)\n", lastmyciatbi - clk + IRQ_OFFSET,
			lastmyciatbi + IRQ_OFFSET);
#endif
#ifdef OLDIRQ
	    myciaint |= 0x80;
#endif
	    myciatbirq = lastmyciatbi + IRQ_OFFSET;
	    if((!myciatbi) || myciatbirq < myciatbi) {
	      maincpu_set_alarm_clk(A_MYCIATB, myciatbirq);
	    } else {
	      maincpu_set_alarm_clk(A_MYCIATB, myciatbi);
	    }
	  }
        }
      }
/*    }*/
    return 0; 
}

int int_myciatod(long offset) {
        register int t, pm;
	
#ifdef DEBUG
if (debugflg)
  printf("MYCIA: TOD timer event (1/10 sec tick), tod=%02x:%02x,%02x.%x\n",
                mycia[CIA_TOD_HR], mycia[CIA_TOD_MIN], mycia[CIA_TOD_SEC],
                mycia[CIA_TOD_TEN]);
#endif
	
        /* set up new int */
        maincpu_set_alarm(A_MYCIATOD, myciatodticks);

        if(!myciatodstopped) {
          /* inc timer */
          t = bcd2byte(mycia[CIA_TOD_TEN]);
          t++;
          mycia[CIA_TOD_TEN] = byte2bcd(t%10);
          if( t>= 10 ) {
            t = bcd2byte(mycia[CIA_TOD_SEC]);
            t++;
            mycia[CIA_TOD_SEC] = byte2bcd(t%60);
            if(t >= 60) {
              t = bcd2byte(mycia[CIA_TOD_MIN]);
              t++;
              mycia[CIA_TOD_MIN] = byte2bcd(t%60);
              if(t>= 60) {
                pm = mycia[CIA_TOD_HR] & 0x80;
                t = bcd2byte(mycia[CIA_TOD_HR]&0x1f);
                if(!t) pm ^= 0x80;      /* toggle am/pm on 0:59->1:00 hr */
                t++;
                t = t%12 | pm;
                mycia[CIA_TOD_HR] = byte2bcd(t);
              }
            }
          }
#ifdef DEBUG
if (debugflg)
  printf("MYCIA: TOD after event :tod=%02x:%02x,%02x.%x\n",
                mycia[CIA_TOD_HR], mycia[CIA_TOD_MIN], mycia[CIA_TOD_SEC],
                mycia[CIA_TOD_TEN]);
#endif
          /* check alarm */
	  check_myciatodalarm();
        }
        return 0;
}

static void check_myciatodalarm(void) {
          if(!memcmp(myciatodalarm, mycia+CIA_TOD_TEN, sizeof(myciatodalarm))) {
            myciaint |= CIA_IM_TOD;
#ifdef OLDIRQ
            if(mycia[CIA_ICR] & CIA_IM_TOD) {
	      maincpu_set_int(I_MYCIATOD, MYCIA_INT);
	      myciaint |= 0x80;	/* EP */
	    }
#else
	    update_myciairq();
#endif
          }
}

/* ------------------------------------------------------------------------- */

static void update_myciata(int offset)
{
    register int latch = (mycia[CIA_TAL] + (mycia[CIA_TAH]<<8)) & 0xffff;

    /*printf("update_myciata(force=%d, stopped=%d)\n",force, myciata_stopped);*/

    myciata = latch;
    myciatau = clk - offset + 1 + INT_OFFSET;
    if(!myciata_stopped) {
        myciatai = clk + myciata - offset + INT_OFFSET;
#ifdef MYCIA_TIMER_DEBUG
printf("update_myciata: maincpu_set_alarm to %ld + %d -> %ld\n", clk, myciata-offset+INT_OFFSET, myciatai);
#endif
	if((!myciatairq) || myciatai < myciatairq) {
          maincpu_set_alarm_clk(A_MYCIATA, myciatai);
	}
    } else {
        maincpu_unset_alarm(A_MYCIATA);
    }
}


static void update_myciatb(int offset)
{
    register int latch = (mycia[CIA_TBL] + (mycia[CIA_TBH] << 8)) & 0xffff;

    if((mycia[CIA_CRB] & 0x60)==0x00) {
/*printf("update: %04x force=%d, latch=%d offset=%d, tbu=%ld, clk=%ld, tb=%d -> ",PC,force,latch,offset,myciatbu, clk, myciatb);*/
	myciatb = latch;
	myciatbu = clk - offset + 1 + INT_OFFSET;
	if(!myciatb_stopped) {
	    myciatbi = clk + myciatb - offset + INT_OFFSET;
#ifdef MYCIA_TIMER_DEBUG
printf("updata_myciatb: maincpu_set_alarm to %ld + %d -> %ld\n", clk, myciatb-offset+INT_OFFSET, myciatbi);
#endif
	    if((!myciatbirq) || myciatbi < myciatbirq)
	      maincpu_set_alarm_clk(A_MYCIATB, myciatbi);	
	} else {
	   maincpu_unset_alarm(A_MYCIATB);	
	}
/*printf("tb=%d\n",myciatb);*/
    } else {		/* count TA underflows */
	myciatb = latch;
    }
}

/* -------------------------------------------------------------------------- */


void mycia_prevent_clk_overflow()
{
    rclk -= PREVENT_CLK_OVERFLOW_SUB;
    myciatau -= PREVENT_CLK_OVERFLOW_SUB;
    myciatai -= PREVENT_CLK_OVERFLOW_SUB;
    myciatairq -= PREVENT_CLK_OVERFLOW_SUB;
    lastmyciatai -= PREVENT_CLK_OVERFLOW_SUB;
    myciatbu -= PREVENT_CLK_OVERFLOW_SUB; 
    myciatbi -= PREVENT_CLK_OVERFLOW_SUB;
    myciatbirq -= PREVENT_CLK_OVERFLOW_SUB;
    lastmyciatbi -= PREVENT_CLK_OVERFLOW_SUB;
    myciardi -= PREVENT_CLK_OVERFLOW_SUB;
}

POST_CIA_FUNCS


