/*
 * $Id: cia.h,v 2.0 1996/04/17 18:45:08 fachat Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * CIA 6526 and VIA 6522 register number declarations.
 * Interrupt signals are defined in vmachine.h.
 *
 * Written by
 *    Jouko Valta  (jopi@stekt.oulu.fi)
 *
 *
 * $Log: cia.h,v $
 * Revision 2.0  1996/04/17 18:45:08  fachat
 * Changed all these P_PORT_* stuff to CIA_*, which is much clearer
 * and doesn't interfere with 6522 VIA defines...
 *
 * Revision 1.3  1995/06/28  19:40:56  jopi
 * Never count on the documents...
 *
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * Defines for registers
 *
 * Revision 1.1  1994/08/10  17:46:41  jopi
 * Initial revision
 *
 *
 */

#ifndef X64_CIA_H
#define X64_CIA_H

#define byte2bcd(byte)  ((16*((byte)/10)+((byte)%10))&0xff)
#define bcd2byte(bcd)   ((10*((bcd&0xf0)/16)+((bcd)&0xf))&0xff)

/*
 * MOS 6526 registers
 */
#define CIA_PRA		0  /* Port A */
#define CIA_PRB		1  /* Port B */
#define CIA_DDRA	2  /* Data direction register for port A */
#define CIA_DDRB	3  /* Data direction register for port B */

#define CIA_TAL		4  /* Timer A low */
#define CIA_TAH		5  /* Timer A high */
#define CIA_TBL		6  /* Timer B low */
#define CIA_TBH		7  /* Timer B high */

#define CIA_TOD_TEN	8  /* Time Of Day clock 1/10 s */
#define CIA_TOD_SEC	9  /* Time Of Day clock sec */
#define CIA_TOD_MIN	10 /* Time Of Day clock min */
#define CIA_TOD_HR	11 /* Time Of Day clock hour */

#define CIA_SDR		12 /* Serial port shift register */
#define CIA_ICR		13 /* Interrupt control register */
#define CIA_CRA		14 /* Control register A */
#define CIA_CRB		15 /* Control register B */


/*
 * Interrupt Masks
 */


/* MOS 6526 */
#define CIA_IM_SET	0x80	/* Control Bit */
#define CIA_IM_TA	1	/* Timer A underflow */
#define CIA_IM_TB	2	/* Timer B underflow */
#define CIA_IM_TOD	4	/* TOD Clock Alarm */
#define CIA_IM_SP	8	/* Shift Register completion */
#define CIA_IM_FLG	16	/* Handshake */


#endif  /* X64_CIA_H */
