
/*
 * cia1.c
 * This file is generated from cia-tmpl.c and cia1.h,
 * Do not edit!
 */
/*
 * $Id: cia-tmpl.c,v 1.1 1997/05/22 19:40:10 fachat Exp $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * Written by
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Fixes by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 * $Log: cia-tmpl.c,v $
 * Revision 1.1  1997/05/22 19:40:10  fachat
 * Initial revision
 *
 */

/* 
 * 15mar97 a.fachat
 * new interrupt handling, hopefully according to the specs now.
 * All interrupts (note: not timer events (i.e. alarms)) are put
 * into one interrupt flag, I_CIA1FL.
 * if an interrupt condition changes, the function (i.e. cpp macro)
 * update_cia1irq() id called, that checks the IRQ line state.
 * This is now possible, as ettore has decoupled A_* alarm events
 * from I_* interrupts for performance reasons.
 */


#define	OLDIRQ

#define	cia1icr	cia1[CIA_ICR]

#define update_cia1irq() 						\
	do { 								\
	  if( cia1int & cia1icr & 0x7f ) cia1int |= 0x80; 		\
    	  maincpu_set_int(I_CIA1FL, cia1int & 0x80 ? IK_IRQ : 0); 	\
	} while(0)

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "extern.h"
#include "vmachine.h"
#include "extern.h"
#include "cpu.h"
#include "cia.h"


	
#include "interrupt.h"

/* #define CIA1_TIMER_DEBUG */

#define	STORE_OFFSET 1 /* store_cia called one cycle after actual store */
#define	INT_OFFSET   1 /* offset given to int_cia is this amount to large */
#define	IRQ_OFFSET   4 /* IRQ line goes this # cycles low after ICR flag set */

/* global */

BYTE cia1[16];
extern BYTE keyarr[KBD_ROWS];

/* local functions */

static void update_cia1ta ( int );
static void update_cia1tb ( int );
static void check_cia1todalarm( void );
int cia1tbx ( int );

/*
 * Local variables
 */


static int cia1int;	/* Interrupt Flag register for cia 1 */

static int cia1ta;	/* value of cia1 timer A at last update */
static int cia1tb;	/* value of cia1 timer B at last update */

static CLOCK rclk;	/* real clock = clk-offset */
static CLOCK cia1tau;	/* time when cia1 timer A is updated */
static CLOCK cia1tai;	/* clock when next IRQ is expected */
static CLOCK cia1tairq;/* clk when to set IRQ line low */
static CLOCK lastcia1tai;
static CLOCK cia1tbu;	/* time when cia1 timer B is updated */
static CLOCK cia1tbi;	/* clock when next IRQ is expected */
static CLOCK cia1tbirq;/* clk when to set IRQ line low */
static CLOCK lastcia1tbi;

static CLOCK cia1rdi;	/* time of last ICR read */

static int cia1ta_stopped = 1;	/* timer stopped */
static int cia1tb_stopped = 1;	/* timer stopped */

static int cia1sr_bits;	/* number of bits still to send * 2 */

static BYTE cia1todalarm[4];
static BYTE cia1todlatch[4];
static char cia1todstopped;
static char cia1todlatched;
static int cia1todticks = 100000;          /* approx. a 1/10 sec. */

/* ------------------------------------------------------------------------- */
/* CIA1 */



static int readta(void) {
	register int tmp;

	if(cia1ta_stopped) 
	  return (cia1ta==0) ? cia1[CIA_TAL]+256*cia1[CIA_TAH] : cia1ta; 

	tmp= (clk-cia1tau + (cia1ta+1)) % (cia1ta+1);

	if(tmp==cia1ta) tmp = cia1ta;
	else tmp = cia1ta - tmp;

	return tmp;
}

static int readtb(void) {
	register int tmp;

	if(cia1tb_stopped) 
	  return (cia1tb==0) ? cia1[CIA_TBL]+256*cia1[CIA_TBH] : cia1tb; 

	tmp= (clk-cia1tbu + (cia1tb+1)) % (cia1tb+1);

	if(tmp==cia1tb) tmp = cia1tb;
	else tmp = cia1tb - tmp;

	return tmp;
}


void    reset_cia1(void)
{
	int i;
	for(i=0;i<16;i++) cia1[i] = 0;

	/* FIXME: Is this actually correct? */
	cia1[CIA_TAL] = cia1[CIA_TBL] = 1;
	
	cia1sr_bits = 0;
	cia1ta_stopped = 0;
	cia1tb_stopped = 0;
	cia1ta = 0xffff;
	cia1tb = 0xffff;
	cia1tau = clk;
	cia1tbu = clk;

	cia1tbi = 0;
	cia1tbirq = 0;
	lastcia1tbi = 0;
	maincpu_unset_alarm(A_CIA1TB);
	cia1tai = 0;
	cia1tairq = 0;
	lastcia1tai = 0;
	maincpu_unset_alarm(A_CIA1TA);

	cia1rdi = 0;

        memset(cia1todalarm, 0, sizeof(cia1todalarm));
        cia1todlatched = 0;
        cia1todstopped = 0;
        maincpu_set_alarm(A_CIA1TOD, cia1todticks);

	cia1int = 0;
	update_cia1irq();
}


void    store_cia1(ADDRESS addr, BYTE byte)
{
    addr &= 0xf;
       
 
#ifdef DEBUG
    if (addr >=4 /*debugflg*/)
	printf("store cia1[%x] %x\n", (int) addr, (int) byte);
#endif

    if(cia1tbi && cia1tbi<clk-STORE_OFFSET) int_cia1tb(clk-cia1tbi);
    if(cia1tai && cia1tai<clk-STORE_OFFSET) int_cia1ta(clk-cia1tai);

    rclk = clk - STORE_OFFSET;

    switch (addr) {

      case CIA_PRA: /* port A */
	
      case CIA_DDRA: 
	cia1[addr] = byte;
	break;
      case CIA_PRB: /* port B */
      case CIA_DDRB: 
	cia1[addr] = byte;
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 *
	 * REAL:  TOD register + (wallclock - cia1todrel)
	 * VIRT:  TOD register + (cycles - begin)/cycles_per_sec
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
        if(cia1[CIA_CRB] & 0x80) cia1todalarm[addr-CIA_TOD_TEN]=byte;
        else {
          if(addr == CIA_TOD_TEN) cia1todstopped = 0;
          if(addr == CIA_TOD_HR) cia1todstopped = 1;
          cia1[addr] = byte;
        }
	check_cia1todalarm();
        break;

      case CIA_SDR: /* Serial Port output buffer */
	cia1[addr] = byte;
	if((cia1[CIA_CRA] & 0x40) && (cia1sr_bits < 16)) cia1sr_bits += 16;
	break;

	/* Interrupts */

      case CIA_ICR: /* Interrupt Control Register */
#if defined (CIA1_TIMER_DEBUG)
        if(debugflg) printf ("CIA1 set CIA_ICR: 0x%x\n", byte);
#endif
	if(byte & CIA_IM_SET) {
	  cia1[addr] |= byte;
	} else {
	  cia1[addr] &= ~byte;
	}
#ifdef OLDIRQ
	maincpu_set_int(I_CIA1TA, (cia1int & cia1[CIA_ICR] & CIA_IM_TA) ? IK_IRQ : 0 );
	if(!(cia1int & cia1[CIA_ICR] & CIA_IM_TA)) cia1tairq = 0;
	if((!cia1tairq) || (cia1tai < cia1tairq)) maincpu_set_alarm_clk(A_CIA1TA, cia1tai);
	maincpu_set_int(I_CIA1TB, (cia1int & cia1[CIA_ICR] & CIA_IM_TB) ? IK_IRQ : 0 );
	if(!(cia1int & cia1[CIA_ICR] & CIA_IM_TB)) cia1tbirq = 0;
	/* TODO: this is a bad hack (tm)! */
	if((!(cia1[CIA_CRB] & 0x60)) && ((!cia1tbirq) || (cia1tbi < cia1tbirq))) maincpu_set_alarm_clk(A_CIA1TB, cia1tbi);
	maincpu_set_int(I_CIA1SP, (cia1int & cia1[CIA_ICR] & CIA_IM_SP) ? IK_IRQ : 0 );
	maincpu_set_int(I_CIA1TOD,(cia1int & cia1[CIA_ICR] & CIA_IM_TOD)? IK_IRQ : 0 );
	/* EP */
	if (cia1int & cia1[CIA_ICR])
	    cia1int |= 0x80;
	break;
#else
	update_cia1irq();
	if(!(cia1int & cia1[CIA_ICR] & CIA_IM_TA)) cia1tairq = 0;
	if((!cia1tairq) || (cia1tai < cia1tairq)) maincpu_set_alarm_clk(A_CIA1TA, cia1tai);
	if(!(cia1int & cia1[CIA_ICR] & CIA_IM_TB)) cia1tbirq = 0;
	/* TODO: this is a bad hack (tm)! */
	if((!(cia1[CIA_CRB] & 0x60)) && ((!cia1tbirq) || (cia1tbi < cia1tbirq))) maincpu_set_alarm_clk(A_CIA1TB, cia1tbi);
#endif

	/* Control */

      case CIA_CRA: /* control register A */
#if defined (CIA1_TIMER_DEBUG)
        printf ("CIA1 set CIA_CRA: 0x%x\n", byte);
#endif

	/* bit 7 tod frequency */
	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1ta(STORE_OFFSET + INT_OFFSET -1);
	    if(!cia1ta_stopped) {
	      	cia1tau = clk;
	      	cia1tai = clk + cia1ta;
		if((!cia1tairq) || (cia1tairq>cia1tai)) {
		    maincpu_set_alarm(A_CIA1TA, cia1ta);
		}
	    }
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
            /* stop timer */
#if defined (CIA1_TIMER_DEBUG)
	    if(debugflg) printf ("CIA1 Timer A stopped\n");
#endif
	    if(!cia1ta_stopped) {
	    	cia1ta = readta(); 	/* set cia1ta to actual value */
		cia1ta_stopped = 1;
		cia1tai = 0;
		if(cia1tairq) {
		    maincpu_set_alarm_clk(A_CIA1TA, cia1tairq);
		} else {
		    maincpu_unset_alarm(A_CIA1TA);
		}
	    }
	} else {
#if defined (CIA1_TIMER_DEBUG)
	    printf ("CIA1 Timer A restarted, cia1ta=%d\n", cia1ta);
#endif
	    if(cia1ta_stopped) {
		cia1tau = clk;
		cia1ta_stopped = 0;
                update_cia1ta(STORE_OFFSET + INT_OFFSET -1); 
		cia1tai = clk + cia1ta;
		if((!cia1tairq) || (cia1tairq>cia1tai))
		    maincpu_set_alarm(A_CIA1TA, cia1ta);
	    }
	}

	cia1[addr] = byte & 0xef;	/* remove strobe */
	break;

      case CIA_CRB: /* control register B */
#if defined (CIA1_TIMER_DEBUG)
        if(1 /*debugflg*/) printf ("CIA1 set CIA_CRB: 0x%x, stopped=%d\n", 
				byte, cia1tb_stopped);
#endif
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1tb(STORE_OFFSET + INT_OFFSET -1);
	    if(!cia1tb_stopped) {
	      	cia1tbu = clk;
	      	cia1tbi = clk + cia1tb;
		if((!cia1tbirq) || (cia1tbirq>cia1tbi))
		    maincpu_set_alarm(A_CIA1TB, cia1tb);
	    }
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x60) || !(byte & 1)) {
	    if(!cia1tb_stopped) {
#if defined (CIA1_TIMER_DEBUG)
 	    	if(debugflg) printf ("CIA1 Timer B stopped\n");
#endif
		cia1tb = readtb(); 	/* set cia1tb to actual value */
		cia1tb_stopped = 1;
		cia1tbi = 0;
		if(cia1tbirq) 
		   maincpu_set_alarm_clk(A_CIA1TB, cia1tbirq);
		else
		   maincpu_unset_alarm(A_CIA1TB);
	    }
	} else {
	    if(cia1tb_stopped) {
#if defined (CIA1_TIMER_DEBUG)
	    	if(debugflg) printf ("CIA1 Timer B restarted\n");
#endif
		cia1tbu = clk;
		cia1tb_stopped = 0;
                update_cia1tb(STORE_OFFSET + INT_OFFSET -1);
		cia1tbi = clk + cia1tb;
/*printf("start cia1tb: clk = %ld, cia1tb =%04x, cia1tbirq =%ld, cia1tbi=%ld\n",
			clk, cia1tb, cia1tbirq, cia1tbi);*/
		if((!cia1tbirq) || (cia1tbirq>cia1tbi))
		    maincpu_set_alarm(A_CIA1TB, cia1tb);
	    }
	}

	cia1[addr] = byte & 0xef;	/* remove strobe */
	break;
      default:
	cia1[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_cia1(ADDRESS addr)
{
    addr &= 0xf;
    
#ifdef DEBUG
    if (addr < 4 /*debugflg*/) printf("read cia1[%x], clk=%ld\n", addr, clk);
#endif
   
/*    if(addr == CIA_ICR) cia1rdi = clk;*/
 
    if(cia1tbi && cia1tbi<clk) int_cia1tb(clk-cia1tbi);
    if(cia1tai && cia1tai<clk) int_cia1ta(clk-cia1tai);

    switch (addr) {

      case CIA_PRA: /* port A */

	{
	    BYTE val = cia1[CIA_PRA] | ~cia1[CIA_DDRA];
	    BYTE msk = (cia1[CIA_PRB] | ~cia1[CIA_DDRB]) & ~joy[1];
	    int m, i;

	    /* FIXME: this is sloooow! */
	    for (m = 0x1; m <= 0x80; m <<= 1) {
		if (!(msk & m)) {
		    for (i = 0; i < KBD_ROWS; i++)
			if (keyarr[i] & m) {
			    val &= ~(1 << i);
			    break;
			}
		}
	    }
	    return val & ~joy[2];
	}
	break;

      case CIA_PRB: /* port B */

	{
	    BYTE val = ~cia1[CIA_DDRB];
	    BYTE msk = (cia1[CIA_PRA] | ~cia1[CIA_DDRA]) & ~joy[2];
	    BYTE m;
	    int i;
	    
	    for (m = 0x1, i = 0; i < KBD_ROWS; m <<= 1, i++)
		if (!(msk & m))
		    val &= ~keyarr[i];
	    return (val | (cia1[CIA_PRB] & cia1[CIA_DDRB])) & ~joy[1];
	}
	break;
	/* Timers */

      case CIA_TAL: /* timer A low */
	return (readta() & 0xff);

      case CIA_TAH: /* timer A high */
	return ((readta() >> 8) & 0xff);

      case CIA_TBL: /* timer B low */
	return readtb() & 0xff;

      case CIA_TBH: /* timer B high */
	return (readtb() >> 8) & 0xff;

	/*
	 * TOD clock is latched by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this latching is different from the input one.
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
        if(!cia1todlatched) memcpy(cia1todlatch, cia1+CIA_TOD_TEN, sizeof(cia1todlatch));
        if(addr == CIA_TOD_TEN) cia1todlatched = 0;
        if(addr == CIA_TOD_HR) cia1todlatched = 1;
        return cia1[addr];

      case CIA_SDR: /* Serial Port Shift Register */
#ifdef DEBUG
	/* little hack .... */
	{ int i; printf("\nmaincpu_ints:");
	for(i=0;i<NUMOFINT;i++) {
	  printf(" %d",maincpu_int_status.pending_int[i]);
	}
	printf("\n"); }
#endif
	return (cia1[addr]);

	/* Interrupts */

      case CIA_ICR: /* Interrupt Flag Register */
	{
	    BYTE    t = cia1int;

	    cia1rdi = clk;
	    /* if ( t & cia1[CIA_ICR] ) t |= 0x80; EP (removed) */

#ifdef DEBUG
	    if(debugflg) 
		printf("CIA1 read intfl gives cia1int=%02x -> %02x @"
		   	" PC = %04x, sr_bits=%d, clk=%ld, ta=%d, tb=%d\n", 
			cia1int, t, PC, cia1sr_bits, clk, readta(), readtb());
#endif
	    cia1int = 0;
	    cia1tairq = 0;
	    cia1tbirq = 0;
	    if(cia1tai) maincpu_set_alarm_clk(A_CIA1TA, cia1tai);
	    if(cia1tbi) maincpu_set_alarm_clk(A_CIA1TB, cia1tbi);
#ifdef OLDIRQ	    
	    maincpu_set_int(I_CIA1TA, 0);
	    maincpu_set_int(I_CIA1TB, 0);
	    maincpu_set_int(I_CIA1SP, 0);
	    maincpu_set_int(I_CIA1TOD,0);
#else
	    update_cia1irq();
#endif
	    return (t);
	}
    }  /* switch */

    return (cia1[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_cia1ta(long offset)
{

    if(offset<0) return 0; 	/* ???????? */

    rclk = clk - offset;

#ifdef DEBUG
    if (debugflg)
	printf("cia1 timer A event\n");
#endif

#ifdef CIA1_TIMER_DEBUG
printf("int_cia1ta: rclk=%ld-%ld=%ld: cia1tai=%ld, lastatai=%ld, tau=%ld, tairq=%ld\n",
			clk, offset, clk-offset,cia1tai,lastcia1tai, cia1tau,
			cia1tairq);
#endif

    if(rclk == cia1tairq) { 
#ifdef OLDIRQ
      if(cia1[CIA_ICR] & CIA_IM_TA) {
	maincpu_set_int(I_CIA1TA, IK_IRQ);
	cia1int |= 0x80;	/* EP */
      }
#else
      cia1int |= CIA_IM_TA;
      update_cia1irq();
#endif
      cia1tairq = 0;
      if(cia1tai) {

#ifdef CIA1_TIMER_DEBUG
printf("int_cia1ta: maincpu_set_alarm(%ld -> %ld)\n",cia1tai - clk, cia1tai);
#endif
	maincpu_set_alarm_clk(A_CIA1TA, cia1tai);
      } else {
	maincpu_unset_alarm(A_CIA1TA);
      }
    } 

    if(rclk == cia1tai) {

      lastcia1tai = cia1tai;

      /* update timer */

      if (cia1[CIA_CRA] & 8) {	/* one-shot mode */
#if defined (CIA1_TIMER_DEBUG)
        if(debugflg) printf ("CIA1 Timer A event -- one-shot mode: next int won't happen\n");
#endif
	cia1ta = cia1[CIA_TAL]+256*cia1[CIA_TAH];
	cia1ta_stopped = 1;
	cia1tai = 0;
	if(cia1tairq) 
	  maincpu_set_alarm_clk(A_CIA1TA, cia1tairq);
	else
	  maincpu_unset_alarm(A_CIA1TA);
	cia1[CIA_CRA] &= ~1;
      } else {		/* continuous mode */
	update_cia1ta(offset);
      }

      /* check SR */

      if(cia1[CIA_CRA] & 0x40) {
        if(cia1sr_bits) {
  	  if(!(--cia1sr_bits)) {
	    cia1int |= CIA_IM_SP;
#ifdef OLDIRQ
	    /* TODO: this sure is another delayed IRQ... */
            if(cia1[CIA_ICR] & CIA_IM_SP) {
	      maincpu_set_int(I_CIA1SP, IK_IRQ);
	      cia1int |= 0x80;	/* EP */
	    }
#else
	    update_cia1irq();
#endif
	  }
        } 
      }

      /* set IRQ line (delayed, if necessary) */

      if(1 /*cia1rdi != rclk-STORE_OFFSET*/) {	/* not cleared by reading ICR */
#ifdef OLDIRQ
        cia1int |= CIA_IM_TA;
#endif
        if(cia1[CIA_ICR] & CIA_IM_TA) {  /* timer bitmask set? */
#ifdef CIA1_TIMER_DEBUG
printf("clk=%ld: lastcia1tai=%ld\n",clk, lastcia1tai);
#endif
	  if((clk-lastcia1tai) > IRQ_OFFSET) {
#ifdef OLDIRQ
	    maincpu_set_int(I_CIA1TA, IK_IRQ);
	    cia1int |= 0x80;	/* EP */
#else
	    cia1int |= CIA_IM_TA;
	    update_cia1irq();
#endif
	    cia1tairq = 0;
	  } else {
#ifdef CIA1_TIMER_DEBUG
printf("int_cia1ta: main_cpu_set_alarm(%ld -> %ld)\n", lastcia1tai - clk + IRQ_OFFSET,
			lastcia1tai + IRQ_OFFSET);
#endif
#ifdef OLDIRQ
	    cia1int |= 0x80;
#endif
	    cia1tairq = lastcia1tai + IRQ_OFFSET;
	    if((!cia1tai) || cia1tairq < cia1tai) {
	      maincpu_set_alarm_clk(A_CIA1TA, cia1tairq);
	    } else {
	      maincpu_set_alarm_clk(A_CIA1TA, cia1tai);
	    }
	  }
	}
      }

    }
      /* count timer A underflow pulses */

      if ((cia1[CIA_CRB] & 0x61) == 0x41) {
	if (--cia1tb < 0) {	/* is this "=" ok? (AF) */
	    cia1tbx(offset); /*int_cia1tb(0); */
	}
      }

    return 0; 
}


/*
 * Timer B can run in 2 (4) modes
 * cia1[f] & 0x60 == 0x00   count system 02 pulses
 * cia1[f] & 0x60 == 0x40   count timer A underflows
 * cia1[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


int    int_cia1tb(long offset)
{

    if(offset<0) {
	return 0; 	/* ???????? */
    }

    rclk = clk - offset;

#ifdef DEBUG
    if (debugflg)
	printf("CIA1 timer B event, clk=%ld, offset=%ld, tbi=%ld, latch=%d,"
		"PC=%04x, rline=%d\n",
		 clk, offset, cia1tbi, cia1[CIA_TBL]+256*cia1[CIA_TBH],
		PC, rasterline);
#endif

#ifdef CIA1_TIMER_DEBUG
printf("int_cia1tb: clk=%ld-%ld: cia1tbi=%ld\n",clk, offset,cia1tbi);
#endif

/*    lastcia1tbi = cia1tbirq;*/

    if(rclk == cia1tbirq) {
#ifdef OLDIRQ
        if(cia1[CIA_ICR] & CIA_IM_TB) {
	  maincpu_set_int(I_CIA1TB, IK_IRQ);
	  cia1int |= 0x80;	/* EP */
	}
#else
	cia1int |= CIA_IM_TB;
	update_cia1irq();
#endif
	cia1tbirq = 0;

	if(cia1tbi) {
#ifdef CIA1_TIMER_DEBUG
printf("int-cia1tb: maincpu_set_alarm(%ld)\n",cia1tbi - clk);
#endif
	  maincpu_set_alarm_clk(A_CIA1TB, cia1tbi);
	} else {
	  maincpu_unset_alarm(A_CIA1TB);
	}
    } 

    /* FIXME: hack to allow int_cia1tb() to be called from int_cia1ta() */
    if(rclk == cia1tbi) {
	return cia1tbx(offset);
    }

    return 0;
}

int cia1tbx(int offset) {

      lastcia1tbi = rclk; /*cia1tbi;*/
/*printf("tbx: clk=%ld, lastcia1tbi=%ld, icr=%02x\n", clk, lastcia1tbi, cia1[CIA_ICR]);*/

      /* update timer */

      if (cia1[CIA_CRB] & 8) {	/* one-shot mode */
#if defined (CIA1_TIMER_DEBUG)
        printf ("CIA1 Timer B event -- one-shot mode: next int won't happen\n");
#endif
	cia1tb = cia1[CIA_TBL]+256*cia1[CIA_TBH];
	cia1tb_stopped = 1;
	cia1tbi = 0;
	if(cia1tbirq) 
	  maincpu_set_alarm_clk(A_CIA1TB, cia1tbirq);
	else
	  maincpu_unset_alarm(A_CIA1TB);
	cia1[CIA_CRB] &= ~1;	/* CRB run bit */
      } else {		/* continuous mode */
	update_cia1tb(offset);
      }

      /* set IRQ line (delayed, if necessary) */

      if(1 /*cia1rdi != rclk-STORE_OFFSET*/) {
#ifdef OLDIRQ
        cia1int |= CIA_IM_TB;
#endif
        if(cia1[CIA_ICR] & CIA_IM_TB) {
#ifdef CIA1_TIMER_DEBUG
printf("clk=%ld: lastcia1tbi=%ld\n",clk, lastcia1tbi);
#endif
	  if((clk-lastcia1tbi) > IRQ_OFFSET) {
#ifdef OLDIRQ
	    maincpu_set_int(I_CIA1TB, IK_IRQ);
	    cia1int |= 0x80;	/* EP */
#else
	    cia1int |= CIA_IM_TB;
	    update_cia1irq();
#endif
	    cia1tbirq = 0;
	  } else {
#ifdef CIA1_TIMER_DEBUG
printf("int_cia1tb: maincpu_set_alarm(%ld -> %ld)\n", lastcia1tbi - clk + IRQ_OFFSET,
			lastcia1tbi + IRQ_OFFSET);
#endif
#ifdef OLDIRQ
	    cia1int |= 0x80;
#endif
	    cia1tbirq = lastcia1tbi + IRQ_OFFSET;
	    if((!cia1tbi) || cia1tbirq < cia1tbi) {
	      maincpu_set_alarm_clk(A_CIA1TB, cia1tbirq);
	    } else {
	      maincpu_set_alarm_clk(A_CIA1TB, cia1tbi);
	    }
	  }
        }
      }
/*    }*/
    return 0; 
}

int int_cia1tod(long offset) {
        register int t, pm;
	
#ifdef DEBUG
if (debugflg)
  printf("CIA1: TOD timer event (1/10 sec tick), tod=%02x:%02x,%02x.%x\n",
                cia1[CIA_TOD_HR], cia1[CIA_TOD_MIN], cia1[CIA_TOD_SEC],
                cia1[CIA_TOD_TEN]);
#endif
	
        /* set up new int */
        maincpu_set_alarm(A_CIA1TOD, cia1todticks);

        if(!cia1todstopped) {
          /* inc timer */
          t = bcd2byte(cia1[CIA_TOD_TEN]);
          t++;
          cia1[CIA_TOD_TEN] = byte2bcd(t%10);
          if( t>= 10 ) {
            t = bcd2byte(cia1[CIA_TOD_SEC]);
            t++;
            cia1[CIA_TOD_SEC] = byte2bcd(t%60);
            if(t >= 60) {
              t = bcd2byte(cia1[CIA_TOD_MIN]);
              t++;
              cia1[CIA_TOD_MIN] = byte2bcd(t%60);
              if(t>= 60) {
                pm = cia1[CIA_TOD_HR] & 0x80;
                t = bcd2byte(cia1[CIA_TOD_HR]&0x1f);
                if(!t) pm ^= 0x80;      /* toggle am/pm on 0:59->1:00 hr */
                t++;
                t = t%12 | pm;
                cia1[CIA_TOD_HR] = byte2bcd(t);
              }
            }
          }
#ifdef DEBUG
if (debugflg)
  printf("CIA1: TOD after event :tod=%02x:%02x,%02x.%x\n",
                cia1[CIA_TOD_HR], cia1[CIA_TOD_MIN], cia1[CIA_TOD_SEC],
                cia1[CIA_TOD_TEN]);
#endif
          /* check alarm */
	  check_cia1todalarm();
        }
        return 0;
}

static void check_cia1todalarm(void) {
          if(!memcmp(cia1todalarm, cia1+CIA_TOD_TEN, sizeof(cia1todalarm))) {
            cia1int |= CIA_IM_TOD;
#ifdef OLDIRQ
            if(cia1[CIA_ICR] & CIA_IM_TOD) {
	      maincpu_set_int(I_CIA1TOD, IK_IRQ);
	      cia1int |= 0x80;	/* EP */
	    }
#else
	    update_cia1irq();
#endif
          }
}

/* ------------------------------------------------------------------------- */

static void update_cia1ta(int offset)
{
    register int latch = (cia1[CIA_TAL] + (cia1[CIA_TAH]<<8)) & 0xffff;

    /*printf("update_cia1ta(force=%d, stopped=%d)\n",force, cia1ta_stopped);*/

    cia1ta = latch;
    cia1tau = clk - offset + 1 + INT_OFFSET;
    if(!cia1ta_stopped) {
        cia1tai = clk + cia1ta - offset + INT_OFFSET;
#ifdef CIA1_TIMER_DEBUG
printf("update_cia1ta: maincpu_set_alarm to %ld + %d -> %ld\n", clk, cia1ta-offset+INT_OFFSET, cia1tai);
#endif
	if((!cia1tairq) || cia1tai < cia1tairq) {
          maincpu_set_alarm_clk(A_CIA1TA, cia1tai);
	}
    } else {
        maincpu_unset_alarm(A_CIA1TA);
    }
}


static void update_cia1tb(int offset)
{
    register int latch = (cia1[CIA_TBL] + (cia1[CIA_TBH] << 8)) & 0xffff;

    if((cia1[CIA_CRB] & 0x60)==0x00) {
/*printf("update: %04x force=%d, latch=%d offset=%d, tbu=%ld, clk=%ld, tb=%d -> ",PC,force,latch,offset,cia1tbu, clk, cia1tb);*/
	cia1tb = latch;
	cia1tbu = clk - offset + 1 + INT_OFFSET;
	if(!cia1tb_stopped) {
	    cia1tbi = clk + cia1tb - offset + INT_OFFSET;
#ifdef CIA1_TIMER_DEBUG
printf("updata_cia1tb: maincpu_set_alarm to %ld + %d -> %ld\n", clk, cia1tb-offset+INT_OFFSET, cia1tbi);
#endif
	    if((!cia1tbirq) || cia1tbi < cia1tbirq)
	      maincpu_set_alarm_clk(A_CIA1TB, cia1tbi);	
	} else {
	   maincpu_unset_alarm(A_CIA1TB);	
	}
/*printf("tb=%d\n",cia1tb);*/
    } else {		/* count TA underflows */
	cia1tb = latch;
    }
}

/* -------------------------------------------------------------------------- */


void cia1_prevent_clk_overflow()
{
    rclk -= PREVENT_CLK_OVERFLOW_SUB;
    cia1tau -= PREVENT_CLK_OVERFLOW_SUB;
    cia1tai -= PREVENT_CLK_OVERFLOW_SUB;
    cia1tairq -= PREVENT_CLK_OVERFLOW_SUB;
    lastcia1tai -= PREVENT_CLK_OVERFLOW_SUB;
    cia1tbu -= PREVENT_CLK_OVERFLOW_SUB; 
    cia1tbi -= PREVENT_CLK_OVERFLOW_SUB;
    cia1tbirq -= PREVENT_CLK_OVERFLOW_SUB;
    lastcia1tbi -= PREVENT_CLK_OVERFLOW_SUB;
    cia1rdi -= PREVENT_CLK_OVERFLOW_SUB;
}



int     show_keyarr(void)
{
    int     i, j;
    for (j = 0; j < KBD_ROWS; j++) {
	printf("%d:", j);
	for (i = 0x80; i; i >>= 1)
	    printf(" %d", keyarr[j] & i);
	printf("\n");
    }
    return (0);
}



