
/*
 * cia2.c
 * This file is generated from cia-tmpl.c and cia2.h,
 * Do not edit!
 */
/*
 * $Id: cia-tmpl.c,v 1.1 1997/05/22 19:40:10 fachat Exp $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * Written by
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Fixes by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 * $Log: cia-tmpl.c,v $
 * Revision 1.1  1997/05/22 19:40:10  fachat
 * Initial revision
 *
 */

/* 
 * 15mar97 a.fachat
 * new interrupt handling, hopefully according to the specs now.
 * All interrupts (note: not timer events (i.e. alarms)) are put
 * into one interrupt flag, I_CIA2FL.
 * if an interrupt condition changes, the function (i.e. cpp macro)
 * update_cia2irq() id called, that checks the IRQ line state.
 * This is now possible, as ettore has decoupled A_* alarm events
 * from I_* interrupts for performance reasons.
 */


#define	OLDIRQ

#define	cia2icr	cia2[CIA_ICR]

#define update_cia2irq() 						\
	do { 								\
	  if( cia2int & cia2icr & 0x7f ) cia2int |= 0x80; 		\
    	  maincpu_set_int(I_CIA2FL, cia2int & 0x80 ? IK_NMI : 0); 	\
	} while(0)

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "extern.h"
#include "vmachine.h"
#include "extern.h"
#include "cpu.h"
#include "cia.h"

#include "true1541.h"
	
#include "interrupt.h"

/* #define CIA2_TIMER_DEBUG */

#define	STORE_OFFSET 1 /* store_cia called one cycle after actual store */
#define	INT_OFFSET   1 /* offset given to int_cia is this amount to large */
#define	IRQ_OFFSET   4 /* IRQ line goes this # cycles low after ICR flag set */

/* global */

BYTE cia2[16];
extern BYTE keyarr[KBD_ROWS];

/* local functions */

static void update_cia2ta ( int );
static void update_cia2tb ( int );
static void check_cia2todalarm( void );
int cia2tbx ( int );

/*
 * Local variables
 */


static int cia2int;	/* Interrupt Flag register for cia 1 */

static int cia2ta;	/* value of cia2 timer A at last update */
static int cia2tb;	/* value of cia2 timer B at last update */

static CLOCK rclk;	/* real clock = clk-offset */
static CLOCK cia2tau;	/* time when cia2 timer A is updated */
static CLOCK cia2tai;	/* clock when next IRQ is expected */
static CLOCK cia2tairq;/* clk when to set IRQ line low */
static CLOCK lastcia2tai;
static CLOCK cia2tbu;	/* time when cia2 timer B is updated */
static CLOCK cia2tbi;	/* clock when next IRQ is expected */
static CLOCK cia2tbirq;/* clk when to set IRQ line low */
static CLOCK lastcia2tbi;

static CLOCK cia2rdi;	/* time of last ICR read */

static int cia2ta_stopped = 1;	/* timer stopped */
static int cia2tb_stopped = 1;	/* timer stopped */

static int cia2sr_bits;	/* number of bits still to send * 2 */

static BYTE cia2todalarm[4];
static BYTE cia2todlatch[4];
static char cia2todstopped;
static char cia2todlatched;
static int cia2todticks = 100000;          /* approx. a 1/10 sec. */

/* ------------------------------------------------------------------------- */
/* CIA2 */



static int readta(void) {
	register int tmp;

	if(cia2ta_stopped) 
	  return (cia2ta==0) ? cia2[CIA_TAL]+256*cia2[CIA_TAH] : cia2ta; 

	tmp= (clk-cia2tau + (cia2ta+1)) % (cia2ta+1);

	if(tmp==cia2ta) tmp = cia2ta;
	else tmp = cia2ta - tmp;

	return tmp;
}

static int readtb(void) {
	register int tmp;

	if(cia2tb_stopped) 
	  return (cia2tb==0) ? cia2[CIA_TBL]+256*cia2[CIA_TBH] : cia2tb; 

	tmp= (clk-cia2tbu + (cia2tb+1)) % (cia2tb+1);

	if(tmp==cia2tb) tmp = cia2tb;
	else tmp = cia2tb - tmp;

	return tmp;
}


void    reset_cia2(void)
{
	int i;
	for(i=0;i<16;i++) cia2[i] = 0;

	/* FIXME: Is this actually correct? */
	cia2[CIA_TAL] = cia2[CIA_TBL] = 1;
	
	cia2sr_bits = 0;
	cia2ta_stopped = 0;
	cia2tb_stopped = 0;
	cia2ta = 0xffff;
	cia2tb = 0xffff;
	cia2tau = clk;
	cia2tbu = clk;

	cia2tbi = 0;
	cia2tbirq = 0;
	lastcia2tbi = 0;
	maincpu_unset_alarm(A_CIA2TB);
	cia2tai = 0;
	cia2tairq = 0;
	lastcia2tai = 0;
	maincpu_unset_alarm(A_CIA2TA);

	cia2rdi = 0;

        memset(cia2todalarm, 0, sizeof(cia2todalarm));
        cia2todlatched = 0;
        cia2todstopped = 0;
        maincpu_set_alarm(A_CIA2TOD, cia2todticks);

	cia2int = 0;
	update_cia2irq();
}


void    store_cia2(ADDRESS addr, BYTE byte)
{
    addr &= 0xf;
       
 
#ifdef DEBUG
    if (addr >=4 /*debugflg*/)
	printf("store cia2[%x] %x\n", (int) addr, (int) byte);
#endif

    if(cia2tbi && cia2tbi<clk-STORE_OFFSET) int_cia2tb(clk-cia2tbi);
    if(cia2tai && cia2tai<clk-STORE_OFFSET) int_cia2ta(clk-cia2tai);

    rclk = clk - STORE_OFFSET;

    switch (addr) {

      case CIA_PRA: /* port A */
	
      case CIA_DDRA: 

	if (cia2[addr] != byte)
	    serial_bus_cpu_write(byte | ~cia2[CIA_DDRA]);
	{
	    static BYTE cia2pa = 0;
	    BYTE tmp;
	    extern int vbank;
	    cia2[addr] = byte;
	    tmp = cia2[CIA_PRA] | ~cia2[CIA_DDRA];
	    if(tmp != cia2pa) {
	      cia2pa = tmp;
	      vbank = ((cia2pa /*cia2[CIA_PRA]*/ & 3) ^ 3) << 14;
	      screen_init();
	    }
	}
	break;
      case CIA_PRB: /* port B */
      case CIA_DDRB: 
	cia2[addr] = byte;
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 *
	 * REAL:  TOD register + (wallclock - cia2todrel)
	 * VIRT:  TOD register + (cycles - begin)/cycles_per_sec
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
        if(cia2[CIA_CRB] & 0x80) cia2todalarm[addr-CIA_TOD_TEN]=byte;
        else {
          if(addr == CIA_TOD_TEN) cia2todstopped = 0;
          if(addr == CIA_TOD_HR) cia2todstopped = 1;
          cia2[addr] = byte;
        }
	check_cia2todalarm();
        break;

      case CIA_SDR: /* Serial Port output buffer */
	cia2[addr] = byte;
	if((cia2[CIA_CRA] & 0x40) && (cia2sr_bits < 16)) cia2sr_bits += 16;
	break;

	/* Interrupts */

      case CIA_ICR: /* Interrupt Control Register */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("CIA2 set CIA_ICR: 0x%x\n", byte);
#endif
	if(byte & CIA_IM_SET) {
	  cia2[addr] |= byte;
	} else {
	  cia2[addr] &= ~byte;
	}
#ifdef OLDIRQ
	maincpu_set_int(I_CIA2TA, (cia2int & cia2[CIA_ICR] & CIA_IM_TA) ? IK_NMI : 0 );
	if(!(cia2int & cia2[CIA_ICR] & CIA_IM_TA)) cia2tairq = 0;
	if((!cia2tairq) || (cia2tai < cia2tairq)) maincpu_set_alarm_clk(A_CIA2TA, cia2tai);
	maincpu_set_int(I_CIA2TB, (cia2int & cia2[CIA_ICR] & CIA_IM_TB) ? IK_NMI : 0 );
	if(!(cia2int & cia2[CIA_ICR] & CIA_IM_TB)) cia2tbirq = 0;
	/* TODO: this is a bad hack (tm)! */
	if((!(cia2[CIA_CRB] & 0x60)) && ((!cia2tbirq) || (cia2tbi < cia2tbirq))) maincpu_set_alarm_clk(A_CIA2TB, cia2tbi);
	maincpu_set_int(I_CIA2SP, (cia2int & cia2[CIA_ICR] & CIA_IM_SP) ? IK_NMI : 0 );
	maincpu_set_int(I_CIA2TOD,(cia2int & cia2[CIA_ICR] & CIA_IM_TOD)? IK_NMI : 0 );
	/* EP */
	if (cia2int & cia2[CIA_ICR])
	    cia2int |= 0x80;
	break;
#else
	update_cia2irq();
	if(!(cia2int & cia2[CIA_ICR] & CIA_IM_TA)) cia2tairq = 0;
	if((!cia2tairq) || (cia2tai < cia2tairq)) maincpu_set_alarm_clk(A_CIA2TA, cia2tai);
	if(!(cia2int & cia2[CIA_ICR] & CIA_IM_TB)) cia2tbirq = 0;
	/* TODO: this is a bad hack (tm)! */
	if((!(cia2[CIA_CRB] & 0x60)) && ((!cia2tbirq) || (cia2tbi < cia2tbirq))) maincpu_set_alarm_clk(A_CIA2TB, cia2tbi);
#endif

	/* Control */

      case CIA_CRA: /* control register A */
#if defined (CIA2_TIMER_DEBUG)
        printf ("CIA2 set CIA_CRA: 0x%x\n", byte);
#endif

	/* bit 7 tod frequency */
	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia2ta(STORE_OFFSET + INT_OFFSET -1);
	    if(!cia2ta_stopped) {
	      	cia2tau = clk;
	      	cia2tai = clk + cia2ta;
		if((!cia2tairq) || (cia2tairq>cia2tai)) {
		    maincpu_set_alarm(A_CIA2TA, cia2ta);
		}
	    }
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
            /* stop timer */
#if defined (CIA2_TIMER_DEBUG)
	    if(debugflg) printf ("CIA2 Timer A stopped\n");
#endif
	    if(!cia2ta_stopped) {
	    	cia2ta = readta(); 	/* set cia2ta to actual value */
		cia2ta_stopped = 1;
		cia2tai = 0;
		if(cia2tairq) {
		    maincpu_set_alarm_clk(A_CIA2TA, cia2tairq);
		} else {
		    maincpu_unset_alarm(A_CIA2TA);
		}
	    }
	} else {
#if defined (CIA2_TIMER_DEBUG)
	    printf ("CIA2 Timer A restarted, cia2ta=%d\n", cia2ta);
#endif
	    if(cia2ta_stopped) {
		cia2tau = clk;
		cia2ta_stopped = 0;
                update_cia2ta(STORE_OFFSET + INT_OFFSET -1); 
		cia2tai = clk + cia2ta;
		if((!cia2tairq) || (cia2tairq>cia2tai))
		    maincpu_set_alarm(A_CIA2TA, cia2ta);
	    }
	}

	cia2[addr] = byte & 0xef;	/* remove strobe */
	break;

      case CIA_CRB: /* control register B */
#if defined (CIA2_TIMER_DEBUG)
        if(1 /*debugflg*/) printf ("CIA2 set CIA_CRB: 0x%x, stopped=%d\n", 
				byte, cia2tb_stopped);
#endif
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia2tb(STORE_OFFSET + INT_OFFSET -1);
	    if(!cia2tb_stopped) {
	      	cia2tbu = clk;
	      	cia2tbi = clk + cia2tb;
		if((!cia2tbirq) || (cia2tbirq>cia2tbi))
		    maincpu_set_alarm(A_CIA2TB, cia2tb);
	    }
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x60) || !(byte & 1)) {
	    if(!cia2tb_stopped) {
#if defined (CIA2_TIMER_DEBUG)
 	    	if(debugflg) printf ("CIA2 Timer B stopped\n");
#endif
		cia2tb = readtb(); 	/* set cia2tb to actual value */
		cia2tb_stopped = 1;
		cia2tbi = 0;
		if(cia2tbirq) 
		   maincpu_set_alarm_clk(A_CIA2TB, cia2tbirq);
		else
		   maincpu_unset_alarm(A_CIA2TB);
	    }
	} else {
	    if(cia2tb_stopped) {
#if defined (CIA2_TIMER_DEBUG)
	    	if(debugflg) printf ("CIA2 Timer B restarted\n");
#endif
		cia2tbu = clk;
		cia2tb_stopped = 0;
                update_cia2tb(STORE_OFFSET + INT_OFFSET -1);
		cia2tbi = clk + cia2tb;
/*printf("start cia2tb: clk = %ld, cia2tb =%04x, cia2tbirq =%ld, cia2tbi=%ld\n",
			clk, cia2tb, cia2tbirq, cia2tbi);*/
		if((!cia2tbirq) || (cia2tbirq>cia2tbi))
		    maincpu_set_alarm(A_CIA2TB, cia2tb);
	    }
	}

	cia2[addr] = byte & 0xef;	/* remove strobe */
	break;
      default:
	cia2[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_cia2(ADDRESS addr)
{
    addr &= 0xf;
    
#ifdef DEBUG
    if (addr < 4 /*debugflg*/) printf("read cia2[%x], clk=%ld\n", addr, clk);
#endif
   
/*    if(addr == CIA_ICR) cia2rdi = clk;*/
 
    if(cia2tbi && cia2tbi<clk) int_cia2tb(clk-cia2tbi);
    if(cia2tai && cia2tai<clk) int_cia2ta(clk-cia2tai);

    switch (addr) {

      case CIA_PRA: /* port A */
	return ((cia2[CIA_PRA] & cia2[CIA_DDRA]) | ((0x3f | serial_bus_cpu_read(	)) & ~cia2[CIA_DDRA]));
	break;

      case CIA_PRB: /* port B */
	return ((cia2[CIA_PRB] & cia2[CIA_DDRB]) | (0xff & ~cia2[CIA_DDRB]));
	break;
	/* Timers */

      case CIA_TAL: /* timer A low */
	return (readta() & 0xff);

      case CIA_TAH: /* timer A high */
	return ((readta() >> 8) & 0xff);

      case CIA_TBL: /* timer B low */
	return readtb() & 0xff;

      case CIA_TBH: /* timer B high */
	return (readtb() >> 8) & 0xff;

	/*
	 * TOD clock is latched by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this latching is different from the input one.
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
        if(!cia2todlatched) memcpy(cia2todlatch, cia2+CIA_TOD_TEN, sizeof(cia2todlatch));
        if(addr == CIA_TOD_TEN) cia2todlatched = 0;
        if(addr == CIA_TOD_HR) cia2todlatched = 1;
        return cia2[addr];

      case CIA_SDR: /* Serial Port Shift Register */
#ifdef DEBUG
	/* little hack .... */
	{ int i; printf("\nmaincpu_ints:");
	for(i=0;i<NUMOFINT;i++) {
	  printf(" %d",maincpu_int_status.pending_int[i]);
	}
	printf("\n"); }
#endif
	return (cia2[addr]);

	/* Interrupts */

      case CIA_ICR: /* Interrupt Flag Register */
	{
	    BYTE    t = cia2int;

	    cia2rdi = clk;
	    /* if ( t & cia2[CIA_ICR] ) t |= 0x80; EP (removed) */

#ifdef DEBUG
	    if(debugflg) 
		printf("CIA2 read intfl gives cia2int=%02x -> %02x @"
		   	" PC = %04x, sr_bits=%d, clk=%ld, ta=%d, tb=%d\n", 
			cia2int, t, PC, cia2sr_bits, clk, readta(), readtb());
#endif
	    cia2int = 0;
	    cia2tairq = 0;
	    cia2tbirq = 0;
	    if(cia2tai) maincpu_set_alarm_clk(A_CIA2TA, cia2tai);
	    if(cia2tbi) maincpu_set_alarm_clk(A_CIA2TB, cia2tbi);
#ifdef OLDIRQ	    
	    maincpu_set_int(I_CIA2TA, 0);
	    maincpu_set_int(I_CIA2TB, 0);
	    maincpu_set_int(I_CIA2SP, 0);
	    maincpu_set_int(I_CIA2TOD,0);
#else
	    update_cia2irq();
#endif
	    return (t);
	}
    }  /* switch */

    return (cia2[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_cia2ta(long offset)
{

    if(offset<0) return 0; 	/* ???????? */

    rclk = clk - offset;

#ifdef DEBUG
    if (debugflg)
	printf("cia2 timer A event\n");
#endif

#ifdef CIA2_TIMER_DEBUG
printf("int_cia2ta: rclk=%ld-%ld=%ld: cia2tai=%ld, lastatai=%ld, tau=%ld, tairq=%ld\n",
			clk, offset, clk-offset,cia2tai,lastcia2tai, cia2tau,
			cia2tairq);
#endif

    if(rclk == cia2tairq) { 
#ifdef OLDIRQ
      if(cia2[CIA_ICR] & CIA_IM_TA) {
	maincpu_set_int(I_CIA2TA, IK_NMI);
	cia2int |= 0x80;	/* EP */
      }
#else
      cia2int |= CIA_IM_TA;
      update_cia2irq();
#endif
      cia2tairq = 0;
      if(cia2tai) {

#ifdef CIA2_TIMER_DEBUG
printf("int_cia2ta: maincpu_set_alarm(%ld -> %ld)\n",cia2tai - clk, cia2tai);
#endif
	maincpu_set_alarm_clk(A_CIA2TA, cia2tai);
      } else {
	maincpu_unset_alarm(A_CIA2TA);
      }
    } 

    if(rclk == cia2tai) {

      lastcia2tai = cia2tai;

      /* update timer */

      if (cia2[CIA_CRA] & 8) {	/* one-shot mode */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("CIA2 Timer A event -- one-shot mode: next int won't happen\n");
#endif
	cia2ta = cia2[CIA_TAL]+256*cia2[CIA_TAH];
	cia2ta_stopped = 1;
	cia2tai = 0;
	if(cia2tairq) 
	  maincpu_set_alarm_clk(A_CIA2TA, cia2tairq);
	else
	  maincpu_unset_alarm(A_CIA2TA);
	cia2[CIA_CRA] &= ~1;
      } else {		/* continuous mode */
	update_cia2ta(offset);
      }

      /* check SR */

      if(cia2[CIA_CRA] & 0x40) {
        if(cia2sr_bits) {
  	  if(!(--cia2sr_bits)) {
	    cia2int |= CIA_IM_SP;
#ifdef OLDIRQ
	    /* TODO: this sure is another delayed IRQ... */
            if(cia2[CIA_ICR] & CIA_IM_SP) {
	      maincpu_set_int(I_CIA2SP, IK_NMI);
	      cia2int |= 0x80;	/* EP */
	    }
#else
	    update_cia2irq();
#endif
	  }
        } 
      }

      /* set IRQ line (delayed, if necessary) */

      if(1 /*cia2rdi != rclk-STORE_OFFSET*/) {	/* not cleared by reading ICR */
#ifdef OLDIRQ
        cia2int |= CIA_IM_TA;
#endif
        if(cia2[CIA_ICR] & CIA_IM_TA) {  /* timer bitmask set? */
#ifdef CIA2_TIMER_DEBUG
printf("clk=%ld: lastcia2tai=%ld\n",clk, lastcia2tai);
#endif
	  if((clk-lastcia2tai) > IRQ_OFFSET) {
#ifdef OLDIRQ
	    maincpu_set_int(I_CIA2TA, IK_NMI);
	    cia2int |= 0x80;	/* EP */
#else
	    cia2int |= CIA_IM_TA;
	    update_cia2irq();
#endif
	    cia2tairq = 0;
	  } else {
#ifdef CIA2_TIMER_DEBUG
printf("int_cia2ta: main_cpu_set_alarm(%ld -> %ld)\n", lastcia2tai - clk + IRQ_OFFSET,
			lastcia2tai + IRQ_OFFSET);
#endif
#ifdef OLDIRQ
	    cia2int |= 0x80;
#endif
	    cia2tairq = lastcia2tai + IRQ_OFFSET;
	    if((!cia2tai) || cia2tairq < cia2tai) {
	      maincpu_set_alarm_clk(A_CIA2TA, cia2tairq);
	    } else {
	      maincpu_set_alarm_clk(A_CIA2TA, cia2tai);
	    }
	  }
	}
      }

    }
      /* count timer A underflow pulses */

      if ((cia2[CIA_CRB] & 0x61) == 0x41) {
	if (--cia2tb < 0) {	/* is this "=" ok? (AF) */
	    cia2tbx(offset); /*int_cia2tb(0); */
	}
      }

    return 0; 
}


/*
 * Timer B can run in 2 (4) modes
 * cia2[f] & 0x60 == 0x00   count system 02 pulses
 * cia2[f] & 0x60 == 0x40   count timer A underflows
 * cia2[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


int    int_cia2tb(long offset)
{

    if(offset<0) {
	return 0; 	/* ???????? */
    }

    rclk = clk - offset;

#ifdef DEBUG
    if (debugflg)
	printf("CIA2 timer B event, clk=%ld, offset=%ld, tbi=%ld, latch=%d,"
		"PC=%04x, rline=%d\n",
		 clk, offset, cia2tbi, cia2[CIA_TBL]+256*cia2[CIA_TBH],
		PC, rasterline);
#endif

#ifdef CIA2_TIMER_DEBUG
printf("int_cia2tb: clk=%ld-%ld: cia2tbi=%ld\n",clk, offset,cia2tbi);
#endif

/*    lastcia2tbi = cia2tbirq;*/

    if(rclk == cia2tbirq) {
#ifdef OLDIRQ
        if(cia2[CIA_ICR] & CIA_IM_TB) {
	  maincpu_set_int(I_CIA2TB, IK_NMI);
	  cia2int |= 0x80;	/* EP */
	}
#else
	cia2int |= CIA_IM_TB;
	update_cia2irq();
#endif
	cia2tbirq = 0;

	if(cia2tbi) {
#ifdef CIA2_TIMER_DEBUG
printf("int-cia2tb: maincpu_set_alarm(%ld)\n",cia2tbi - clk);
#endif
	  maincpu_set_alarm_clk(A_CIA2TB, cia2tbi);
	} else {
	  maincpu_unset_alarm(A_CIA2TB);
	}
    } 

    /* FIXME: hack to allow int_cia2tb() to be called from int_cia2ta() */
    if(rclk == cia2tbi) {
	return cia2tbx(offset);
    }

    return 0;
}

int cia2tbx(int offset) {

      lastcia2tbi = rclk; /*cia2tbi;*/
/*printf("tbx: clk=%ld, lastcia2tbi=%ld, icr=%02x\n", clk, lastcia2tbi, cia2[CIA_ICR]);*/

      /* update timer */

      if (cia2[CIA_CRB] & 8) {	/* one-shot mode */
#if defined (CIA2_TIMER_DEBUG)
        printf ("CIA2 Timer B event -- one-shot mode: next int won't happen\n");
#endif
	cia2tb = cia2[CIA_TBL]+256*cia2[CIA_TBH];
	cia2tb_stopped = 1;
	cia2tbi = 0;
	if(cia2tbirq) 
	  maincpu_set_alarm_clk(A_CIA2TB, cia2tbirq);
	else
	  maincpu_unset_alarm(A_CIA2TB);
	cia2[CIA_CRB] &= ~1;	/* CRB run bit */
      } else {		/* continuous mode */
	update_cia2tb(offset);
      }

      /* set IRQ line (delayed, if necessary) */

      if(1 /*cia2rdi != rclk-STORE_OFFSET*/) {
#ifdef OLDIRQ
        cia2int |= CIA_IM_TB;
#endif
        if(cia2[CIA_ICR] & CIA_IM_TB) {
#ifdef CIA2_TIMER_DEBUG
printf("clk=%ld: lastcia2tbi=%ld\n",clk, lastcia2tbi);
#endif
	  if((clk-lastcia2tbi) > IRQ_OFFSET) {
#ifdef OLDIRQ
	    maincpu_set_int(I_CIA2TB, IK_NMI);
	    cia2int |= 0x80;	/* EP */
#else
	    cia2int |= CIA_IM_TB;
	    update_cia2irq();
#endif
	    cia2tbirq = 0;
	  } else {
#ifdef CIA2_TIMER_DEBUG
printf("int_cia2tb: maincpu_set_alarm(%ld -> %ld)\n", lastcia2tbi - clk + IRQ_OFFSET,
			lastcia2tbi + IRQ_OFFSET);
#endif
#ifdef OLDIRQ
	    cia2int |= 0x80;
#endif
	    cia2tbirq = lastcia2tbi + IRQ_OFFSET;
	    if((!cia2tbi) || cia2tbirq < cia2tbi) {
	      maincpu_set_alarm_clk(A_CIA2TB, cia2tbirq);
	    } else {
	      maincpu_set_alarm_clk(A_CIA2TB, cia2tbi);
	    }
	  }
        }
      }
/*    }*/
    return 0; 
}

int int_cia2tod(long offset) {
        register int t, pm;
	
#ifdef DEBUG
if (debugflg)
  printf("CIA2: TOD timer event (1/10 sec tick), tod=%02x:%02x,%02x.%x\n",
                cia2[CIA_TOD_HR], cia2[CIA_TOD_MIN], cia2[CIA_TOD_SEC],
                cia2[CIA_TOD_TEN]);
#endif
	
        /* set up new int */
        maincpu_set_alarm(A_CIA2TOD, cia2todticks);

        if(!cia2todstopped) {
          /* inc timer */
          t = bcd2byte(cia2[CIA_TOD_TEN]);
          t++;
          cia2[CIA_TOD_TEN] = byte2bcd(t%10);
          if( t>= 10 ) {
            t = bcd2byte(cia2[CIA_TOD_SEC]);
            t++;
            cia2[CIA_TOD_SEC] = byte2bcd(t%60);
            if(t >= 60) {
              t = bcd2byte(cia2[CIA_TOD_MIN]);
              t++;
              cia2[CIA_TOD_MIN] = byte2bcd(t%60);
              if(t>= 60) {
                pm = cia2[CIA_TOD_HR] & 0x80;
                t = bcd2byte(cia2[CIA_TOD_HR]&0x1f);
                if(!t) pm ^= 0x80;      /* toggle am/pm on 0:59->1:00 hr */
                t++;
                t = t%12 | pm;
                cia2[CIA_TOD_HR] = byte2bcd(t);
              }
            }
          }
#ifdef DEBUG
if (debugflg)
  printf("CIA2: TOD after event :tod=%02x:%02x,%02x.%x\n",
                cia2[CIA_TOD_HR], cia2[CIA_TOD_MIN], cia2[CIA_TOD_SEC],
                cia2[CIA_TOD_TEN]);
#endif
          /* check alarm */
	  check_cia2todalarm();
        }
        return 0;
}

static void check_cia2todalarm(void) {
          if(!memcmp(cia2todalarm, cia2+CIA_TOD_TEN, sizeof(cia2todalarm))) {
            cia2int |= CIA_IM_TOD;
#ifdef OLDIRQ
            if(cia2[CIA_ICR] & CIA_IM_TOD) {
	      maincpu_set_int(I_CIA2TOD, IK_NMI);
	      cia2int |= 0x80;	/* EP */
	    }
#else
	    update_cia2irq();
#endif
          }
}

/* ------------------------------------------------------------------------- */

static void update_cia2ta(int offset)
{
    register int latch = (cia2[CIA_TAL] + (cia2[CIA_TAH]<<8)) & 0xffff;

    /*printf("update_cia2ta(force=%d, stopped=%d)\n",force, cia2ta_stopped);*/

    cia2ta = latch;
    cia2tau = clk - offset + 1 + INT_OFFSET;
    if(!cia2ta_stopped) {
        cia2tai = clk + cia2ta - offset + INT_OFFSET;
#ifdef CIA2_TIMER_DEBUG
printf("update_cia2ta: maincpu_set_alarm to %ld + %d -> %ld\n", clk, cia2ta-offset+INT_OFFSET, cia2tai);
#endif
	if((!cia2tairq) || cia2tai < cia2tairq) {
          maincpu_set_alarm_clk(A_CIA2TA, cia2tai);
	}
    } else {
        maincpu_unset_alarm(A_CIA2TA);
    }
}


static void update_cia2tb(int offset)
{
    register int latch = (cia2[CIA_TBL] + (cia2[CIA_TBH] << 8)) & 0xffff;

    if((cia2[CIA_CRB] & 0x60)==0x00) {
/*printf("update: %04x force=%d, latch=%d offset=%d, tbu=%ld, clk=%ld, tb=%d -> ",PC,force,latch,offset,cia2tbu, clk, cia2tb);*/
	cia2tb = latch;
	cia2tbu = clk - offset + 1 + INT_OFFSET;
	if(!cia2tb_stopped) {
	    cia2tbi = clk + cia2tb - offset + INT_OFFSET;
#ifdef CIA2_TIMER_DEBUG
printf("updata_cia2tb: maincpu_set_alarm to %ld + %d -> %ld\n", clk, cia2tb-offset+INT_OFFSET, cia2tbi);
#endif
	    if((!cia2tbirq) || cia2tbi < cia2tbirq)
	      maincpu_set_alarm_clk(A_CIA2TB, cia2tbi);	
	} else {
	   maincpu_unset_alarm(A_CIA2TB);	
	}
/*printf("tb=%d\n",cia2tb);*/
    } else {		/* count TA underflows */
	cia2tb = latch;
    }
}

/* -------------------------------------------------------------------------- */


void cia2_prevent_clk_overflow()
{
    rclk -= PREVENT_CLK_OVERFLOW_SUB;
    cia2tau -= PREVENT_CLK_OVERFLOW_SUB;
    cia2tai -= PREVENT_CLK_OVERFLOW_SUB;
    cia2tairq -= PREVENT_CLK_OVERFLOW_SUB;
    lastcia2tai -= PREVENT_CLK_OVERFLOW_SUB;
    cia2tbu -= PREVENT_CLK_OVERFLOW_SUB; 
    cia2tbi -= PREVENT_CLK_OVERFLOW_SUB;
    cia2tbirq -= PREVENT_CLK_OVERFLOW_SUB;
    lastcia2tbi -= PREVENT_CLK_OVERFLOW_SUB;
    cia2rdi -= PREVENT_CLK_OVERFLOW_SUB;
}




