/*
 * $Id: tape.c,v 1.3 1996/07/29 22:59:29 ettore Exp $
 *
 * Emulator Tape Image header and filename handling.
 *
 *
 * This file is built from the "cbmarcs.c" module for fvcbm ver. 2.1,
 * written by Dan Fandrich.
 *
 * Compile this file with "pack structures" compiler flag if not GNU C
 *
 *
 * Written by:
 *   Dan Fandrich  (dan@fch.wimsey.bc.ca)
 *   Jouko Valta   (jopi@stekt.oulu.fi)
 *
 *
 * $Log: tape.c,v $
 * Revision 1.3  1996/07/29 22:59:29  ettore
 * T64 support.
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "types.h"
#include "drive.h"
#include "tape.h"


#ifndef SEEK_SET
#define SEEK_SET 0
typedef long       fpos_t;
#endif

#ifdef __GNUC__
#define PACK __attribute__ ((packed))	/* pack structures on byte boundaries */
#else
#define PACK		/* pack using a compiler switch instead */
#endif


typedef struct {
	BYTE Magic[8] PACK;
	char CbmName[17] PACK;
	BYTE RecordSize PACK;		/* REL file record size */
} X00HDR;


/* Global */
 
extern char *slot_type[];

static const BYTE MagicHeaderP00[8] = "C64File\0";


/*---------------------------------------------------------------------------*/

#if (!defined(GEMDOS) && !defined(__MSDOS__))

#include <sys/stat.h>

long filelength (int handle);


/*
 * Returns the length of an open file in bytes
 */

long filelength(int handle)
{
	struct stat statbuf;

	if (fstat(handle, &statbuf))
		return -1;

	return (statbuf.st_size);
}
#endif  /* UNIX */


/* ------------------------------------------------------------------------- */

/*
 * Check if given filename has valid P00 extension.
 * Return the filetype or -1 if no match.
 */

int  is_pc64name (char *name)
{
    int   t = -1;
    char *p;


    if (name == NULL || (p = strrchr(name, '.')) == NULL || strlen(++p) != 3)
	return (-1);

    if (!isdigit(p[1]) || !isdigit(p[2]))
	return (-1);

    switch (toupper(*p)) {
      case 'D': t = FT_DEL; break;
      case 'S': t = FT_SEQ; break;
      case 'P': t = FT_PRG; break;
      case 'U': t = FT_USR; break;
      case 'R': t = FT_REL; break;
    }

    return (t);
}


/*
 * The read pointer must be positioned at the beginning of the
 * file before calling this routine. (stdin cannot be fseek(2)'ed)
 */

int  read_pc64header (FILE *fd, char *name, int *reclen)
{
    X00HDR  hdr;


    if (fread (&hdr, sizeof(X00HDR), 1, fd) != 1) {
	fprintf (stderr, "\nP00 header read failed.\n");
	return (FD_RDERR);
    }

    if (memcmp (hdr.Magic, MagicHeaderP00, 8) != 0)
	return (FD_BADIMAGE);

    strncpy (name, hdr.CbmName, 16);
    *reclen = hdr.RecordSize;		/* REL file record size */

    return (FD_OK);
}


int  write_pc64header (FILE *fd, char *name, int reclen)
{
    X00HDR  hdr;


    memset (&hdr, 0, sizeof(X00HDR));
    strncpy ((char *)hdr.Magic, (char *)MagicHeaderP00, 8);
    strncpy (hdr.CbmName, name, 16);

    hdr.RecordSize = reclen;		/* REL file record size */

    if (fseek(fd, 0, SEEK_SET) != 0)
	return (FD_WRTERR);

    return
	(fwrite(&hdr, sizeof(X00HDR), 1, fd));
}


/*
 * If the file is a PC64 file, read the actual filename from the header.
 * Otherwise, return the filename converted to appropriate PETSCII
 * representation.
 *
 * This function cannot be used on stdin.
 */

char  *pc_get_cbmname (FILE *fd, char *fsname)
{
    static char cbmname[24];
    char  *p;
    int    type;
    int    reclen;


    type = is_pc64name (fsname);

    fseek(fd, 0, SEEK_SET);

    if (type >= 0 && read_pc64header (fd, cbmname, &reclen) == FD_OK) {
	for (p = cbmname; *p && (BYTE)*p != 0xA0; ++p);
	sprintf (p, ",%s", slot_type[type]);
    }
    else {
	rewind (fd);		/* There is no header */


	/* name */

	if ((p = strrchr(fsname, '/')) == NULL || !p++)
	    p = fsname;

	if (!p || !*p)
	    return (NULL);


	/* Get Basename and extension from FSname */

	strncpy (cbmname, p, 16);

	if ((p = strrchr(fsname, '.')) != NULL && strlen(++p) == 3) {
	    ;
	}

	if (type >= 0) {
	    p = cbmname + strlen(cbmname);
	    sprintf (p, ",%s", slot_type[type]);
	}
	else
	    strcat (cbmname, ",prg");		/* Best default */
    } /* !p00 */


    return (cbmname);
}



/* ------------------------------------------------------------------------- */

/*
 * Read directory
 *
 * P00 is just a regular file with a simple header prepended, so just read the
 * header and return the name and addresses evaluated.
 *
 * This function cannot be used on stdin.
 */


int  DirP00(FILE *fd, char *CbmName, int *start, int *end)
{
    X00HDR  hdr;
    unsigned int  FileLength;
    int     realtype = -1;
    char   *FileExt;
    BYTE    buf[2];


    if (fseek(fd, 0, SEEK_SET) != 0)
	return (FD_RDERR);

    if (fread(&hdr, sizeof(hdr), 1, fd) != 1) {
	return (FD_RDERR);
    }


    /* If archive type is unknown, see if file is REL */

    if (/*(realtype == X00) && */ (hdr.RecordSize > 0))
	realtype = FT_REL;


    switch (realtype) {
	case FT_DEL: FileExt = "DEL"; break;
	case FT_SEQ: FileExt = "SEQ"; break;
	case FT_PRG: FileExt = "PRG"; break;
	case FT_USR: FileExt = "USR"; break;
	case FT_REL: FileExt = "REL"; break;
	default:  FileExt = "???";
    }

    strncpy(CbmName, (char *) hdr.CbmName, 16);
    CbmName[16] = 0;		/* never need this on a good P00 file */

    strcat (CbmName, ",");
    strcat (CbmName, FileExt);


    /* Start and End Addresses */

    FileLength = filelength(fileno(fd)) - sizeof(hdr) -2; /* skip start addr */

    if (fread(buf, 2, 1, fd) != 1) {
	return (FD_RDERR);
    }

    *start= *buf | (buf[1] << 8);
    *end = *start + FileLength;

    return (FD_OK);
}


/* ------------------------------------------------------------------------- */

/*
 * T64 Tape Image handling functions
 */

#define TAPE_HDR_SIZE	64
#define TAPE_DIR_SIZE	32

#define TAPE64_HEADERS	2

static const char *MagicHeaderT64[] = {
  "C64 tape image file",
  "C64S tape file",
  "C64S tape image file" /* EP */
};


/* ------------------------------------------------------------------------- */

/*
 * Check if given filename has valid .T64 extension.
 * Return the filetype or -1 if no match.
 */

int  is_t64name (char *name)
{
    char *p;


    if (name == NULL || (p = strrchr(name, '.')) == NULL || strlen(++p) != 3)
	return (-1);

    if (toupper(*p) == 'T' && p[1] == '6' && p[2] == '4')
	return (1);

    return (-1);
}


/*
 * Check the image for a valid T64 header and return the
 * number of files it carries.
 */

int  check_t64_header (FILE *fd)
{
    BYTE  buf[TAPE_HDR_SIZE];
    int   maxentries = 0;


    if (fseek(fd, 0, SEEK_SET) != 0)
	return (FD_RDERR);

    fread(buf, 1, TAPE_HDR_SIZE, fd);


    if (memcmp(buf, (BYTE *)MagicHeaderT64[0], strlen(MagicHeaderT64[0])) &&
	memcmp(buf, (BYTE *)MagicHeaderT64[1], strlen(MagicHeaderT64[1])) &&
        memcmp(buf, (BYTE *)MagicHeaderT64[2], strlen(MagicHeaderT64[2])))
	return (FD_BADIMAGE);


    if (buf[32] != 0 || buf[33] != 1)
	fprintf (stderr, "Warning: Unknown tape version.\n");

    maxentries = buf[34];
    maxentries |= (buf[35] << 8);

    if (!maxentries) {
	printf("Cannot locate any files on the tape.\n");
	return (FD_BADIMAGE);		/* tapeimage, that is */
    }

    printf("\nTape version %d.%.2d with %2d out of %d files.\n",
	   buf[33], buf[32],
	   (buf[37] << 8) | buf[36],  (buf[35] << 8) | buf[34] );

    return (maxentries);
}

