/*
 * $Id: ui_xaw.h,v 1.1 1997/05/22 21:40:27 ettore Exp $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README file for copyright notice.
 *
 * Written by
 *   Ettore Perazzoli   (ettore@comm2000.it)
 *
 * $Log: ui_xaw.h,v $
 * Revision 1.1  1997/05/22 21:40:27  ettore
 * Initial revision
 *
 */

/* Xaw-based GUI. */

#ifndef _UI_XAW_H
#define _UI_XAW_H

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "types.h"
#include "xdebug.h"

typedef Widget UiWindow;
typedef XtCallbackProc UiMenuCallback;
typedef void (*UiExposureHandler) (unsigned int width, unsigned int height);
typedef enum { Button_None, Button_Close, Button_Ok, Button_Cancel,
	       Button_Yes, Button_No } UiButton;
typedef struct { unsigned short red, green, blue; BYTE dither; } UiColorDef;
typedef XtPointer UiCallbackDataPtr;

extern int UiInit(int *argc, char **argv);
extern int UiInitFinish(void);
extern UiWindow UiOpenCanvasWindow(const char *title, int width, int height,
				   int no_autorepeat,
				   UiExposureHandler exposure_proc,
				   int num_colors,
				   const UiColorDef color_defs[],
				   PIXEL pixel_return[]);
extern void UiResizeCanvasWindow(UiWindow w, int height, int width);
extern void UiMapCanvasWindow(UiWindow w);
extern void UiUnmapCanvasWindow(UiWindow w);
extern Window UiCanvasDrawable(UiWindow w);
extern void UiDisplaySpeed(float percent, float framerate);
extern void UiDisplayPaused(void);
extern void UiDispatchEvents(void);
extern void UiError(const char *format,...);
extern void UiMessage(const char *format,...);
extern void UiShowText(const char *title, const char *text);
extern char *UiFileSelect(const char *title);
extern UiButton UiInputString(const char *title, const char *prompt, char *buf,
			      unsigned int buflen);
extern UiButton UiAskConfirmation(const char *title, const char *text);
extern void UiAutoRepeatOn(void);
extern void UiAutoRepeatOff(void);
extern void UiUpdateMenus(void);

#endif /* !defined (_UI_XAW_H) */
